"""Test the links.streams module."""

# Builtins

# Packages

from phylline.links.streams import StreamLink


LOWER_EVENTS = ['foo,', 'bar,', 'foobar!']
LOWER_BUFFERS = [event.encode('utf-8') for event in LOWER_EVENTS]
HIGHER_EVENTS = ['Hello,', 'world!']
HIGHER_BUFFERS = [event.encode('utf-8') for event in HIGHER_EVENTS]
HIGHER_STREAM = ''.join(HIGHER_EVENTS).encode('utf-8')


def test_stream_link():
    """Exercise StreamLink's interface."""
    # print('Testing Stream Link:')
    stream_link = StreamLink()
    assert repr(stream_link) == '⇌~ StreamLink ~⇌'
    for buffer in LOWER_BUFFERS:
        stream_link.to_read(buffer)
    result = stream_link.read()
    assert result == b'foo,bar,foobar!'
    for buffer in HIGHER_BUFFERS:
        stream_link.write(buffer)
    result = stream_link.to_write()
    assert result == HIGHER_STREAM



"""Test the links.loopback module."""

# Builtins

# Packages

from phylline.links.loopback import BottomLoopbackLink, TopLoopbackLink

from tests.unit.links.streams import LOWER_BUFFERS, LOWER_EVENTS


def test_top_loopback_link():
    """Exercise TopLoopbackLink's interface."""
    # print('Testing Top Loopback Link:')
    top_loopback_link = TopLoopbackLink()
    for event in LOWER_EVENTS:
        top_loopback_link.to_receive(event)
        assert top_loopback_link.has_to_send()
    for (i, event) in enumerate(top_loopback_link.to_send_all()):
        # print('Top Loopback Link echoed event: {}'.format(event))
        assert event.data == LOWER_EVENTS[i]
    top_loopback_link.to_read(LOWER_BUFFERS[0])
    top_loopback_link.to_read(LOWER_BUFFERS[1] + LOWER_BUFFERS[2])
    result = top_loopback_link.to_write()
    # print('Top Loopback Link echoed from stream: {}'.format(result))
    assert result == b'foo,bar,foobar!'


def test_bottom_loopback_link():
    """Exercise BottomLoopbackLink's interface."""
    # print('Testing Bottom Loopback Link:')
    bottom_loopback_link = BottomLoopbackLink()
    for event in LOWER_EVENTS:
        bottom_loopback_link.send(event)
        assert bottom_loopback_link.has_receive()
    for (i, event) in enumerate(bottom_loopback_link.receive_all()):
        # print('Bottom Loopback Link echoed event: {}'.format(event))
        assert event.data == LOWER_EVENTS[i]
    bottom_loopback_link.write(LOWER_BUFFERS[0])
    bottom_loopback_link.write(LOWER_BUFFERS[1] + LOWER_BUFFERS[2])
    result = bottom_loopback_link.read()
    # print('Top Loopback Link echoed from stream: {}'.format(result))
    assert result == b'foo,bar,foobar!'



"""Test the links.events module."""

# Builtins

# Packages

from phylline.links.events import EventLink

from tests.unit.links.streams import HIGHER_EVENTS, LOWER_EVENTS


def test_event_link():
    """Exercise EventLink's interface."""
    # print('Testing Event Link:')
    event_link = EventLink()
    for event in LOWER_EVENTS:
        event_link.to_receive(event)
    assert event_link.has_receive()
    for (i, event) in enumerate(event_link.receive_all()):
        # print('Event Link received from queue: {}'.format(event))
        assert event.data == LOWER_EVENTS[i]
    for event in HIGHER_EVENTS:
        event_link.send(event)
    assert event_link.has_to_send()
    for (i, event) in enumerate(event_link.to_send_all()):
        # print('Event Link sent to queue: {}'.format(event))
        assert event.data == HIGHER_EVENTS[i]



"""Support for nicer logging."""

# Builtins

# Packages


def hex_bytes(buffer, truncate=10):
    """Return a hex string representation of a byte buffer."""
    if truncate is not None and truncate < 0:
        raise ValueError('Invalid truncation length is negative: {}'.format(truncate))

    if len(buffer) == 0:
        return '(0 bytes)'

    truncated_length = 0
    if truncate is not None and len(buffer) > truncate:
        truncated_length = len(buffer) - truncate
        buffer = buffer[:truncate]
    hex_string = ' '.join('0x{:02x}'.format(b) for b in buffer)
    if truncated_length > 0:
        if truncated_length == 1:
            unit = 'byte'
        else:
            unit = 'bytes'
        if hex_string:
            hex_string += ' ...({} more {})'.format(truncated_length, unit)
        else:
            hex_string = ('({} {})'.format(truncated_length, unit))
    return hex_string



"""Base classes and utilities for writing communication links.

Links consist of a pair of processors: one processor to process incoming data from
the layer below and surface any resulting data to the layer above, and another to
process incoming data from the layer above and surface any resulting data to the
layer below. Then a link is a pair of independent Kahn process network nodes
which process data in opposite directions.

Provides interfaces to write byte buffer-based stream links or queue-based
discrete event links.
"""

# Builtins

from abc import abstractmethod

# Packages

from phylline.processors import read, stream_processor, write


class StreamLinkAbove(object):
    """Interface for exposing a StreamLink-like interface for the layer above.

    This interface is analogous to the interface for conn1 of a duplex
    multiprocessing.Pipe.
    """

    @abstractmethod
    def read(self):
        """Return available received bytes, while consuming them."""
        pass

    @abstractmethod
    def write(self, event):
        """Write data on the link."""
        pass


class StreamLinkBelow(object):
    """Interface for exposing a StreamLink-like interface for the layer below.

    This interface is analogous to the interface for conn2 of a duplex
    multiprocessing.Pipe.
    """

    @abstractmethod
    def to_read(self, event):
        """Read data on the link."""
        pass

    @abstractmethod
    def to_write(self):
        """Return available bytes to write from the link, while consuming them."""
        pass


# Link base classes


class StreamLink(StreamLinkBelow, StreamLinkAbove):
    """A link which sends and receives bytes of data over a stream.

    The default behavior is to pass the bytes through in both directions without
    any transformation. To change this behavior, either override the class with
    a custom sender and receiver processor, or (not recommended) provide a custom
    processor upon construction.

    Note that each StreamLink adds a bit of performance overhead - it is better
    to use EventLinks and/or SimpleStreamLinks instead.

    Interface:
    Above: sends and receives bytestrings.
    Below: to_send and to_receive bytestrings.
    """

    def __init__(
        self, name=None, reader_processor=None, writer_processor=None,
        reader_processor_args=(), reader_processor_kwargs={},
        writer_processor_args=(), writer_processor_kwargs={},
    ):
        """Initialize processors."""
        super().__init__()
        self.name = name
        if reader_processor is None:
            reader_processor = self.reader_processor
        self._reader = reader_processor(
            *reader_processor_args, **reader_processor_kwargs
        )
        if writer_processor is None:
            writer_processor = self.writer_processor
        self._writer = writer_processor(
            *writer_processor_args, **writer_processor_kwargs
        )

    def __repr__(self):
        """Return a string representation of the link."""
        return '⇌~ {} ~⇌'.format(self.__class__.__qualname__)

    # Implement StreamLinkAbove

    def read(self):
        """Implement StreamLinkAbove.read."""
        return self._reader.read()

    def write(self, bytes_data):
        """Implement StreamLinkAbove.write."""
        # print('{} writing: {}'.format(self.__class__.__qualname__, bytes_data))
        self._writer.send(bytes_data)

    # Implement StreamLinkBelow

    def to_read(self, bytes_data):
        """Implement StreamLinkBelow.to_read."""
        self._reader.send(bytes_data)

    def to_write(self):
        """Implement StreamLinkBelow.to_write."""
        return self._writer.read()

    # Utilities for writing read and write processors

    def after_read(self, buffer):
        """Enqueue the buffer for the layer above for consumption by that layer.

        This gets overridden to change the receiver's behavior to send the event.
        Note that this gets monkey-patched by other links and link utilities
        which combine links, such as ChunkedStreamLink and AutomaticPipe!
        """
        # print('Stream link after_read: {}'.format(buffer))
        yield from write(buffer)

    def after_write(self, buffer):
        """Enqueue the buffer for the layer below for consumption by that layer.

        This gets overridden to change the receiver's behavior to send the event.
        Note that this gets monkey-patched by other links and link utilities
        which combine links, such as ChunkedStreamLink and AutomaticPipe!
        """
        # print('Stream link after_write: {}'.format(buffer))
        yield from write(buffer)

    # Read and write processors

    @stream_processor
    def reader_processor(self):
        """Stream reader processor."""
        while True:
            buffer = yield from read()
            # print('Stream Link read: {}'.format(buffer))
            yield from self.after_read(buffer)

    @stream_processor
    def writer_processor(self):
        """Stream writer processor."""
        while True:
            buffer = yield from read()
            # print('Stream Link writing: {}'.format(buffer))
            yield from self.after_write(buffer)



"""Base classes and utilities for writing communication links.

Links consist of a pair of processors: one processor to process incoming data from
the layer below and surface any resulting data to the layer above, and another to
process incoming data from the layer above and surface any resulting data to the
layer below. Then a link is a pair of independent Kahn process network nodes
which process data in opposite directions.

Provides interfaces to write byte buffer-based stream links or queue-based
discrete event links.
"""

# Builtins

# Packages

from phylline.links.events import DataEventLink, EventLink, EventLinkAbove, EventLinkBelow
from phylline.links.streams import StreamLink, StreamLinkAbove, StreamLinkBelow
from phylline.processors import event_processor, receive, send
from phylline.processors import read_until, stream_processor, write


class GenericLinkBelow(EventLinkBelow, StreamLinkBelow):
    """Interface for exposing a generic Link interface for the layer below."""

    pass


class GenericLinkAbove(EventLinkAbove, StreamLinkAbove):
    """Interface for exposing a generic Link interface for the layer above."""

    pass


# Utility links


class ChunkedStreamLink(StreamLinkBelow, EventLinkAbove, DataEventLink):
    """Processor which sends and receives delimited chunks of data over a stream.

    Adapts byte buffer-based stream processors to event queue processors by
    discretizing the stream into chunks delimited by a chunk separator in the stream.
    When begin_chunk_separator is enabled, each chunk will be delimited by the
    chunk separator at both the start and end of the chunk; otherwise, it will
    only be delimited by a chunk separator at the end of the chunk.

    Omits empty chunks, which will be neither sent nor received.

    Interface:
    Above: sends and receives bytestrings of chunks.
    Below: to_send and to_receive delimited bytestrings (not following chunk boundaries).
    """

    def __init__(self, name=None, chunk_separator=b'\0', begin_chunk_separator=True):
        """Initialize processors."""
        self._event_link = EventLink(
            sender_processor=self.sender_processor,
            sender_processor_args=(chunk_separator, begin_chunk_separator),
            receiver_event_passthrough=True
        )
        self._event_link.after_receive = self.__after_receive
        self._stream_link = StreamLink(
            reader_processor=self.reader_processor,
            reader_processor_args=(chunk_separator,)
        )
        self._stream_link.after_write = self.__after_write
        self.name = name

    def __repr__(self):
        """Return a string representation of the link."""
        if self.name is not None:
            return '⇌~ {}({}) □⇌'.format(self.__class__.__qualname__, self.name)
        else:
            return '⇌~ {} □⇌'.format(self.__class__.__qualname__)

    # Implement EventLinkAbove

    def receive(self):
        """Implement EventLinkAbove.receive."""
        return self._event_link.receive()

    def has_receive(self):
        """Implement EventLinkAbove.has_receive."""
        return self._event_link.has_receive()

    def send(self, chunk):
        """Implement EventLinkAbove.send."""
        self._event_link.send(chunk)

    # Implement StreamLinkBelow

    def to_read(self, bytes_data):
        """Implement StreamLinkBelow.to_read."""
        self._stream_link.to_read(bytes_data)

    def to_write(self):
        """Implement StreamLinkBelow.to_write."""
        return self._stream_link.to_write()

    # Utilities for writing receive and send processors

    def __after_receive(self, event):
        yield from self.after_receive(event)

    def __after_write(self, event):
        yield from self.after_write(event)

    def after_receive(self, event):
        """Enqueue the event for the layer above for consumption by that layer.

        This gets overridden to change the receiver's behavior to send the event.
        Note that this gets monkey-patched by other links and link utilities
        which combine links, such as ChunkedStreamLink and AutomaticPipe!
        """
        # print('Event link after_receive: {}'.format(event))
        yield from send(event)

    def after_write(self, buffer):
        """Enqueue the buffer for the layer below for consumption by that layer.

        This gets overridden to change the receiver's behavior to send the event.
        Note that this gets monkey-patched by other links and link utilities
        which combine links, such as ChunkedStreamLink and AutomaticPipe!
        """
        # print('Stream link after_write: {}'.format(buffer))
        yield from write(buffer)

    # Receive and send processors

    @stream_processor
    def reader_processor(self, chunk_separator):
        """Stream reader processor."""
        while True:
            buffer = yield from read_until(chunk_separator)
            if len(buffer) == 0:
                continue
            # print(hex_bytes(buffer))
            data_event = self.make_link_data(buffer, 'up', None)
            self._event_link.to_receive(data_event)

    @event_processor
    def sender_processor(self, chunk_separator, begin_chunk_separator):
        """Event sender processor."""
        while True:
            event = yield from receive()
            data_event = self.get_link_data(event, 'down')
            event = data_event.data
            if len(event) == 0:
                continue
            stream_contents = event + chunk_separator
            if begin_chunk_separator:
                stream_contents = chunk_separator + stream_contents
            self._stream_link.write(stream_contents)



"""Test the links.clocked module."""

# Builtins

# Packages

from phylline.links.clocked import DelayedEventLink, LinkClockRequest

from tests.unit.links.streams import HIGHER_EVENTS, LOWER_EVENTS


def assert_clock_request_event_received(delayed_event_link, time):
    """Assert that the DelayedEventLink has exactly one event on each end, a LinkClockRequest."""
    assert delayed_event_link.has_receive()
    received_events = list(delayed_event_link.receive_all())
    assert len(received_events) == 1
    assert isinstance(received_events[0], LinkClockRequest)
    assert received_events[0].requested_time == time


def assert_clock_request_event_to_send(delayed_event_link, time):
    """Assert that the DelayedEventLink has exactly one event on each end, a LinkClockRequest."""
    assert delayed_event_link.has_to_send()
    to_send_events = list(delayed_event_link.to_send_all())
    assert len(to_send_events) == 1
    assert isinstance(to_send_events[0], LinkClockRequest)
    assert to_send_events[0].requested_time == time


def test_delayed_event_link():
    """Exercise DelayedEventLink's timing interface."""
    # print('Testing Delayed Event Link:')
    delayed_event_link = DelayedEventLink()
    # Receiving times by clock updates with LinkClockTime events
    # print('On-time clock test:')
    delayed_event_link.update_clock(0)
    for event in LOWER_EVENTS:
        delayed_event_link.to_receive(event)
    for event in HIGHER_EVENTS:
        delayed_event_link.send(event)
    assert_clock_request_event_received(delayed_event_link, 1)
    delayed_event_link.update_clock(0.5)
    assert not delayed_event_link.has_receive()
    assert not delayed_event_link.has_to_send()
    delayed_event_link.update_clock(0.99)
    assert not delayed_event_link.has_receive()
    assert not delayed_event_link.has_to_send()
    delayed_event_link.update_clock(1.0)
    assert delayed_event_link.has_receive()
    for (i, event) in enumerate(delayed_event_link.receive_all()):
        # print('Event Link received from queue: {}'.format(event))
        assert event.data == LOWER_EVENTS[i]
    assert delayed_event_link.has_to_send()
    for (i, event) in enumerate(delayed_event_link.to_send_all()):
        # print('Event Link sent to queue: {}'.format(event))
        assert event.data == HIGHER_EVENTS[i]

    # print('Late clock test:')
    delayed_event_link.update_clock(0)
    for event in HIGHER_EVENTS:
        delayed_event_link.send(event)
    for event in LOWER_EVENTS:
        delayed_event_link.to_receive(event)
    assert_clock_request_event_to_send(delayed_event_link, 1)
    delayed_event_link.update_clock(1.5)
    assert delayed_event_link.has_receive()
    for (i, event) in enumerate(delayed_event_link.receive_all()):
        # print('Event Link received from queue: {}'.format(event))
        assert event.data == LOWER_EVENTS[i]
    assert delayed_event_link.has_to_send()
    for (i, event) in enumerate(delayed_event_link.to_send_all()):
        # print('Event Link sent to queue: {}'.format(event))
        assert event.data == HIGHER_EVENTS[i]

    # print('Staggered clock test:')
    delayed_event_link.update_clock(0)
    delayed_event_link.to_receive(LOWER_EVENTS[0])
    delayed_event_link.send(HIGHER_EVENTS[0])
    assert_clock_request_event_received(delayed_event_link, 1)
    assert not delayed_event_link.has_to_send()
    delayed_event_link.update_clock(0.5)
    for event in LOWER_EVENTS[1:]:
        delayed_event_link.to_receive(event)
    for event in HIGHER_EVENTS[1:]:
        delayed_event_link.send(event)
    delayed_event_link.update_clock(1.0)
    assert delayed_event_link.has_receive()
    received_events = list(delayed_event_link.receive_all())
    # print(received_events)
    assert len(received_events) == 2
    assert received_events[0].data == LOWER_EVENTS[0]
    assert isinstance(received_events[1], LinkClockRequest)
    assert received_events[1].requested_time == 1.5
    assert delayed_event_link.has_to_send()
    to_send_events = list(delayed_event_link.to_send_all())
    assert len(to_send_events) == 1
    for (i, event) in enumerate(to_send_events):
        # print('Event Link sent to queue: {}'.format(event))
        assert event.data == HIGHER_EVENTS[i]
    delayed_event_link.update_clock(1.5)
    assert delayed_event_link.has_receive()
    received_events = list(delayed_event_link.receive_all())
    assert len(received_events) == 2
    for (i, event) in enumerate(received_events):
        # print('Event Link received from queue: {}'.format(event))
        assert event.data == LOWER_EVENTS[i + 1]
    assert delayed_event_link.has_to_send()
    to_send_events = list(delayed_event_link.to_send_all())
    assert len(to_send_events) == 1
    for (i, event) in enumerate(to_send_events):
        # print('Event Link sent to queue: {}'.format(event))
        assert event.data == HIGHER_EVENTS[i + 1]



"""Test the util.timing module."""

# Builtins

import math
import time

# Packages

from phylline.util.timing import Clock, TimeoutTimer


def test_clock_realtime():
    """Test real-time clock functionality."""
    clock = Clock()
    assert clock.realtime
    old_time = clock.time
    time.sleep(0.1)
    assert clock.time - old_time >= 0.1
    assert clock.time - old_time < 0.2
    clock.update()
    assert clock.time - old_time >= 0.1
    assert clock.time - old_time < 0.2
    time.sleep(0.1)
    assert clock.time - old_time >= 0.2
    assert clock.time - old_time < 0.3
    clock.update(time=0.0)
    assert clock.time - old_time >= 0.2
    assert clock.time - old_time < 0.3
    clock.reset(time=0)
    assert not clock.realtime
    assert clock.time == 0.0


def test_clock_external():
    """Test externally-clocked clock functionality."""
    clock = Clock(time=0.0)
    assert not clock.realtime
    assert clock.time == 0.0
    clock.update()
    assert clock.time == 0.0
    clock.update(time=0.1)
    assert clock.time == 0.1
    clock.update()
    assert clock.time == 0.1
    clock.update(time=0.0)
    assert clock.time == 0.0
    clock.reset()
    assert clock.realtime
    assert clock.time > 0


def assert_timer_stopped(timer):
    """Check whether the timer is stopped."""
    assert not timer.enabled
    assert timer.elapsed is None
    assert not timer.running
    assert not timer.timed_out
    assert timer.remaining is None
    assert repr(timer) == 'TimeoutTimer(stopped timeout={})'.format(timer.timeout)


def assert_timer_halfway(timer):
    """Check whether the timer has correct state when running halfway to timeout."""
    if timer.clock.realtime:
        assert timer.elapsed >= 0.5 * timer.timeout
        assert timer.elapsed < timer.timeout
    else:
        assert math.isclose(timer.elapsed, 0.5 * timer.timeout)
    assert timer.running
    assert not timer.timed_out
    assert timer.remaining > 0
    if timer.clock.realtime:
        assert timer.remaining <= 0.5 * timer.timeout
    else:
        assert math.isclose(timer.remaining, 0.5 * timer.timeout)


def assert_timer_finished(timer):
    """Check whether the timer has correct state after it has timed out."""
    if timer.clock.realtime:
        assert timer.elapsed >= timer.timeout
    else:
        assert timer.elapsed >= timer.timeout or math.isclose(timer.elapsed, timer.timeout)
    assert not timer.running
    assert timer.timed_out
    assert timer.remaining == 0


def assert_timer_running_realtime(timer):
    """Check whether the timer is running correctly in real-time mode."""
    time.sleep(0.5 * timer.timeout)
    assert_timer_halfway(timer)
    time.sleep(0.5 * timer.timeout)
    assert_timer_finished(timer)
    time.sleep(0.5 * timer.timeout)
    assert_timer_finished(timer)


def test_timer_realtime():
    """Test real-time timer functionality."""
    timer = TimeoutTimer(timeout=0.2)
    assert timer.timeout_time is None
    assert timer.elapsed is None
    timer.enabled = True
    assert timer.elapsed is None  # no start time was set due to invalid state change
    timer.enabled = False
    assert_timer_stopped(timer)

    # Basic test
    timer.start()
    assert_timer_running_realtime(timer)

    # Stop-and-start test
    timer.reset_and_stop()
    assert_timer_stopped(timer)
    timer.reset()
    assert_timer_stopped(timer)
    timer.start()
    assert_timer_running_realtime(timer)

    # Reset test
    timer.reset()
    assert_timer_running_realtime(timer)
    timer.reset()
    assert_timer_running_realtime(timer)

    # Start test
    timer.start(timeout=0.1)
    assert_timer_running_realtime(timer)
    timer.reset()
    assert_timer_running_realtime(timer)
    timer.start()
    assert_timer_running_realtime(timer)

    # Externally-clocked mode-switching test
    timer.clock.reset(0.0)
    timer.start(timeout=0.2)
    timer.clock.update(0.1)
    assert_timer_halfway(timer)
    timer.clock.update(0.2)
    assert_timer_finished(timer)


def test_timer_external():
    """Test externally-clocked timer functionality."""
    timer = TimeoutTimer(timeout=0.2, clock=Clock(time=0.0))
    assert timer.timeout_time is None
    assert timer.elapsed is None
    assert_timer_stopped(timer)

    # Basic test
    timer.clock.update(1.0)
    timer.start()
    assert repr(timer) == 'TimeoutTimer(running from 1.0, timeout=0.2)'
    assert timer.timeout_time == 1.2
    timer.clock.update(1.1)
    assert_timer_halfway(timer)
    timer.clock.update(1.2)
    assert_timer_finished(timer)
    timer.clock.update(1.3)
    assert_timer_finished(timer)
    timer.clock.update(0.0)
    timer.reset()
    timer.clock.update(0.1)
    assert_timer_halfway(timer)
    timer.clock.update(0.2)
    assert_timer_finished(timer)

    # Shared-clock test
    timer.clock.update(1.0)
    timer.reset()
    timer_slow = TimeoutTimer(timeout=0.4, clock=timer.clock)
    timer_slow.start()
    timer.clock.update(1.1)
    assert_timer_halfway(timer)
    timer.clock.update(1.2)
    assert_timer_finished(timer)
    assert_timer_halfway(timer_slow)
    timer.clock.update(1.4)
    assert_timer_finished(timer)
    assert_timer_finished(timer_slow)



"""Test the pipes module."""

# Builtins

# Packages

from phylline.links.clocked import DelayedEventLink
from phylline.links.events import EventLink
from phylline.links.links import ChunkedStreamLink
from phylline.pipes import AutomaticPipe, ManualPipe

from tests.unit.links.clocked import assert_clock_request_event_received
from tests.unit.links.links import HIGHER_CHUNKED_STREAM, LOWER_CHUNKED_BUFFERS
from tests.unit.links.streams import HIGHER_BUFFERS, LOWER_BUFFERS


def write_bottom_chunked_buffers(stream_link_below):
    """Write the bottom chunked buffers to the link with StreamLinkBelow."""
    stream_link_below.to_read(LOWER_CHUNKED_BUFFERS[0])
    stream_link_below.to_read(LOWER_CHUNKED_BUFFERS[1] + LOWER_CHUNKED_BUFFERS[2])


def assert_bottom_events(event_link_above):
    """Receive the bottom buffers as events from the link with EventLinkAbove."""
    assert event_link_above.has_receive()
    for (i, event) in enumerate(event_link_above.receive_all()):
        assert event.data == LOWER_BUFFERS[i]


def write_top_events(event_link_above):
    """Receive the top events on the link with EventLinkAbove."""
    for event in HIGHER_BUFFERS:
        event_link_above.send(event)


def test_manual_pipe():
    """Exercise ManualPipe's interface."""
    # print('Testing Piped Event Links with Manual Synchronization:')
    chunked_stream_link = ChunkedStreamLink()
    event_link = EventLink()
    pipe = ManualPipe(chunked_stream_link, event_link)

    # Read/write on links
    write_bottom_chunked_buffers(chunked_stream_link)
    pipe.sync()
    assert_bottom_events(event_link)
    write_top_events(event_link)
    pipe.sync()
    result = chunked_stream_link.to_write()
    # print('Chunked Stream Link wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM

    # Read/write on pipe
    write_bottom_chunked_buffers(pipe)
    pipe.sync()
    assert_bottom_events(pipe)
    write_top_events(pipe)
    pipe.sync()
    result = pipe.to_write()
    # print('Chunked Stream Link wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM


def test_manual_pipe_clocked():
    """Exercise ManualPipe's clock functionality."""
    # print('Testing Piped Clocked Event Links with Manual Synchronization:')
    pipe = ManualPipe(ChunkedStreamLink(), DelayedEventLink())

    # print('On deadline:')
    assert pipe.update_clock(0) is None
    write_bottom_chunked_buffers(pipe)
    assert pipe.sync() is None
    assert_clock_request_event_received(pipe, 1.0)
    assert pipe.update_clock(0.5) is None
    assert not pipe.has_receive()
    assert pipe.update_clock(0.99) is None
    assert not pipe.has_receive()
    assert pipe.update_clock(1.0) is None
    assert_bottom_events(pipe)

    # print('After deadline:')
    assert pipe.update_clock(0) is None
    write_top_events(pipe)
    assert pipe.update_clock(0.5) == 1.0
    assert not pipe.to_write()
    assert pipe.update_clock(0.75) == 1.0
    assert not pipe.to_write()
    assert pipe.update_clock(1.5) is None
    result = pipe.to_write()
    assert result == HIGHER_CHUNKED_STREAM


def test_manual_pipe_composition():
    """Exercise ManualPipe nested composition."""
    # print('Testing Nesting of Piped Event Links with Manual Synchronization:')
    chunked_stream_link = ChunkedStreamLink()
    upper_event_link = EventLink()
    outer_pipe = ManualPipe(
        ManualPipe(chunked_stream_link, EventLink()),
        ManualPipe(EventLink(), upper_event_link)
    )

    # Read/write on links
    write_bottom_chunked_buffers(chunked_stream_link)
    outer_pipe.sync()
    assert_bottom_events(upper_event_link)
    write_top_events(upper_event_link)
    outer_pipe.sync()
    result = chunked_stream_link.to_write()
    # print('Chunked Stream Link wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM

    # Read/write on pipe
    write_bottom_chunked_buffers(outer_pipe)
    outer_pipe.sync()
    assert_bottom_events(outer_pipe)
    write_top_events(outer_pipe)
    outer_pipe.sync()
    result = outer_pipe.to_write()
    # print('Chunked Stream Link wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM


def test_automatic_pipe():
    """Exercise AutomaticPipe's interface."""
    # print('Testing Piped Event Links with Automatic Synchronization:')
    chunked_stream_link = ChunkedStreamLink()
    event_link = EventLink()
    pipe = AutomaticPipe(chunked_stream_link, event_link)

    # Read/write on links
    write_bottom_chunked_buffers(chunked_stream_link)
    assert_bottom_events(event_link)
    write_top_events(event_link)
    result = chunked_stream_link.to_write()
    # print('Chunked Stream Link wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM

    # Read/write on pipe
    write_bottom_chunked_buffers(pipe)
    assert_bottom_events(pipe)
    write_top_events(pipe)
    result = pipe.to_write()
    # print('Chunked Stream Link wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM


def test_automatic_pipe_clocked():
    """Exercise AutomaticPipe's clock functionality."""
    # print('Testing Piped Clocked Event Links with Automatic Synchronization:')
    pipe = AutomaticPipe(ChunkedStreamLink(), DelayedEventLink())
    assert pipe.update_clock(0) is None
    write_bottom_chunked_buffers(pipe)
    assert_clock_request_event_received(pipe, 1.0)
    assert pipe.update_clock(0.5) is None
    assert not pipe.has_receive()
    assert pipe.update_clock(0.99) is None
    assert not pipe.has_receive()
    assert pipe.update_clock(1.0) is None
    assert_bottom_events(pipe)
    # print('Resetting clock...')
    assert pipe.update_clock(0) is None
    # print('Writing top events...')
    write_top_events(pipe)
    # print('Wrote top events!')
    assert pipe.update_clock(0.5) == 1.0
    assert not pipe.to_write()
    assert pipe.update_clock(0.75) == 1.0
    assert not pipe.to_write()
    assert pipe.update_clock(1.5) is None
    result = pipe.to_write()
    assert result == HIGHER_CHUNKED_STREAM


def test_automatic_pipe_composition():
    """Exercise AutomaticPipe nested composition."""
    # print('Testing Nesting of Piped Event Links with Automatic Synchronization:')
    chunked_stream_link = ChunkedStreamLink()
    upper_event_link = EventLink()
    outer_pipe = AutomaticPipe(
        AutomaticPipe(chunked_stream_link, EventLink()),
        AutomaticPipe(EventLink(), upper_event_link)
    )

    # Read/write on links
    write_bottom_chunked_buffers(chunked_stream_link)
    assert_bottom_events(upper_event_link)
    write_top_events(upper_event_link)
    result = chunked_stream_link.to_write()
    # print('Chunked Stream Link wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM

    # Read/write on pipe
    write_bottom_chunked_buffers(outer_pipe)
    assert_bottom_events(outer_pipe)
    write_top_events(outer_pipe)
    result = outer_pipe.to_write()
    # print('Chunked Stream Link wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM



"""Support for specifying interface classes."""

# Builtins

import abc

# Packages


class InterfaceClass(abc.ABCMeta):
    """Metaclass that allows docstring inheritance.

    References:
    https://github.com/sphinx-doc/sphinx/issues/3140#issuecomment-259704637

    """

    def __new__(cls, classname, bases, cls_dict):
        """Initialize class."""
        cls = abc.ABCMeta.__new__(cls, classname, bases, cls_dict)
        mro = cls.__mro__[1:]
        for (name, member) in cls_dict.items():
            if not getattr(member, '__doc__'):
                for base in mro:
                    try:
                        member.__doc__ = getattr(base, name).__doc__
                        break
                    except AttributeError:
                        pass
        return cls


class SetterProperty(object):
    """Data descriptor which is like a property but only supports setting.

    References:
    https://stackoverflow.com/a/17576095

    """

    def __init__(self, func, doc=None):
        """Initialize members."""
        self.func = func
        self.__doc__ = doc if doc is not None else func.__doc__

    def __set__(self, obj, value):
        """Set the value."""
        return self.func(obj, value)



"""Timed execution of tasks."""

# Builtins

import math
import time

# Packages


class Clock(object):
    """Clock which runs in either real-time or on an external clock.

    To use the clock in real-time mode, simply call __init__/reset with no
    arguments. Then everything will use the system clock, and no external times
    need to be input (so that update_clock does not need to be called).
    To use the timer in externally clocked mode, call __init__/reset with the
    current clock time, and call update_clock every time the timer needs to be
    checked. Then everything will use the external clock times given by reset
    """

    def __init__(self, time=None):
        """Initialize members."""
        self._time = time

    def reset(self, time=None):
        """Reset the clock."""
        self._time = time

    def update(self, time=None):
        """Update the clock time.

        Only needs to be caled when the clock runs in externally clocked mode,
        which is done by calling reset/start with the current clock time. If
        called when the clock runs in real-time mode, does nothing.
        """
        if self.realtime:
            return
        if time is None:  # clock in externally-clocked mode, need valid time
            return
        self._time = time

    @property
    def time(self):
        """Clock time."""
        if self.realtime:
            return time.time()
        return self._time

    @property
    def realtime(self):
        """Whether the clock is running in real-time mode."""
        return self._time is None


class TimeoutTimer(object):
    """Timer which counts down to timeout.

    Note that the clock is checked against timeout up to floating-point precision,
    so if the elapsed time is nominally just before the timeout but actually within
    floating-point error of it, the timer will consider itself to have timed out.
    """

    def __init__(self, timeout=0, clock=None):
        """Initialize members."""
        self.enabled = False
        self.timeout = timeout
        self.start_time = None
        if clock is None:
            clock = Clock()
        self.clock = clock

    def start(self, timeout=None):
        """Start the timer from beginning."""
        if timeout is not None:
            self.timeout = timeout
        self.enabled = True
        self.reset()

    def reset(self):
        """Reset the timer."""
        if not self.enabled:
            return
        self.start_time = self.clock.time

    def reset_and_stop(self):
        """Reset the timer and stop it."""
        self.enabled = False
        self.start_time = None

    @property
    def timeout_time(self):
        """Clock time when timeout will occur."""
        if self.start_time is None:
            return None
        return self.start_time + self.timeout

    @property
    def elapsed(self):
        """Amount of time elapsed since start."""
        if not self.enabled:
            return None
        if self.start_time is None:
            return None
        return self.clock.time - self.start_time

    @property
    def running(self):
        """Whether the timer is running to timeout."""
        return (
            self.enabled and (self.elapsed < self.timeout)
            and not math.isclose(self.elapsed, self.timeout)
        )

    @property
    def timed_out(self):
        """Whether timeout has occurred."""
        return self.enabled and (
            self.elapsed >= self.timeout
            or math.isclose(self.elapsed, self.timeout)
        )

    @property
    def remaining(self):
        """Amount of time remaining before timeout."""
        if not self.enabled:
            return None
        duration = self.timeout - self.elapsed
        if self.timed_out:  # check timed_out after duration for real-time correctness
            return 0
        return duration

    def __repr__(self):
        """Return a string representation of the timer."""
        return '{}({}{}{})'.format(
            self.__class__.__qualname__,
            'running ' if self.enabled else 'stopped ',
            'from {}, '.format(self.start_time) if self.start_time is not None else '',
            'timeout={}'.format(self.timeout)
        )



"""Classes for building pipelines of layered links."""

# Builtins
import itertools

# Packages

from phylline.links.clocked import LinkClockRequest, LinkClockTime
from phylline.links.links import GenericLinkAbove, GenericLinkBelow
from phylline.processors import proceed, wait  # Any protocols
from phylline.util.interfaces import SetterProperty
from phylline.util.iterables import make_collection, remove_none


# Pipes


class Pipe(GenericLinkBelow, GenericLinkAbove):
    """Link to join two layers of links together in a pipeline.

    This makes it possible to implement Kahn process networks with links.

    This is just a base class without any implementation of how data is pipelined
    between the layers.

    Note that the pipe only passes events between EventLinks and streams between
    StreamLinks, because there is no standardized way to convert between streams
    and arbitrary events!
    """

    def __init__(self, bottom, top):
        """Initialize members.

        Instantiate the pipe before passing any data to bottom or top, or else
        you will probably lose data!

        If you pass in an iterable of multiple links in bottom and/or top, then
        anything from receive() of any bottom link will be passed to to_read of
        every link in top, and anything from to_send of any top link will
        be passed to send of every link in bottom.
        """
        self.bottom = make_collection(bottom)
        self.top = make_collection(top)
        self.name = None

    def __repr__(self):
        """Return a string representation of the pipeline.

        Left-to-right order is from bottom to top.
        """
        if len(self.bottom) == 1:
            bottom = str(self.bottom[0])
        else:
            bottom = repr(self.bottom)
        if len(self.top) == 1:
            top = str(self.top[0])
        else:
            top = repr(self.top)
        return '[{} | {}]'.format(bottom, top)

    # Utilities for implementing Pipes

    def receive_up(self, event):
        """Pass an event up to all top links which receive events."""
        # print('Passing event up: {}'.format(event))
        for top in self.top:
            try:
                top.to_receive(event)
            except AttributeError:
                pass

    def read_up(self, buffer):
        """Pass a stream buffer up to all top links which read streams."""
        # print('Passing buffer up: {}'.format(buffer))
        for top in self.top:
            try:
                top.to_read(buffer)
            except AttributeError:
                pass

    def send_down(self, event):
        """Pass an event down to all bottom links which send events."""
        # print('Passing event down: {}'.format(event))
        for bottom in self.bottom:
            try:
                bottom.send(event)
            except AttributeError:
                pass

    def write_down(self, buffer):
        """Pass a stream buffer down to all bottom links which write streams."""
        # print('Passing buffer down: {}'.format(buffer))
        for bottom in self.bottom:
            try:
                bottom.write(buffer)
            except AttributeError:
                pass

    # EventLink/StreamLink-like interface

    @SetterProperty
    def after_receive(self, handler):
        """Mimic EventLink.after_receive."""
        for top in self.top:
            if hasattr(top, 'after_receive'):
                top.after_receive = handler

    @SetterProperty
    def after_send(self, handler):
        """Mimic EventLink.after_receive."""
        for bottom in self.bottom:
            if hasattr(bottom, 'after_send'):
                bottom.after_send = handler

    @SetterProperty
    def after_read(self, handler):
        """Mimic StreamLink.after_read."""
        for top in self.top:
            if hasattr(top, 'after_read'):
                top.after_read = handler

    @SetterProperty
    def after_write(self, handler):
        """Mimic StreamLink.after_write."""
        for bottom in self.bottom:
            if hasattr(bottom, 'after_write'):
                bottom.after_write = handler

    # Clocks

    def update_clock(self, time):
        """Update the clock of any ClockedLink and do any necessary processing."""
        for link in itertools.chain(self.bottom, self.top):
            try:
                link.update_clock(time)
            except AttributeError:
                pass

    # Implement GenericLinkBelow

    def to_receive(self, event):
        """Implement EventLinkBelow.to_receive."""
        for bottom in self.bottom:
            try:
                bottom.to_receive(event)
            except AttributeError:
                pass

    def to_send(self):
        """Implement EventLinkBelow.to_send."""
        for bottom in self.bottom:
            try:
                if bottom.has_to_send():
                    return bottom.to_send()
            except AttributeError:
                pass

    def has_to_send(self):
        """Implement EventLinkBelow.has_to_send."""
        for bottom in self.bottom:
            try:
                if bottom.has_to_send():
                    return True
            except AttributeError:
                pass
        return False

    def to_read(self, event):
        """Implement StreamLinkBelow.to_read."""
        for bottom in self.bottom:
            try:
                bottom.to_read(event)
            except AttributeError:
                pass

    def to_write(self):
        """Implement StreamLinkBelow.to_write."""
        all_to_write = b''
        for bottom in self.bottom:
            try:
                if all_to_write == b'':
                    all_to_write = bottom.to_write()
                else:
                    all_to_write += bottom.to_write()
            except AttributeError:
                pass
        return all_to_write

    # Implement GenericLinkAbove

    def receive(self):
        """Implement EventLinkAbove.receive."""
        for top in self.top:
            try:
                if top.has_receive():
                    return top.receive()
            except AttributeError:
                pass
        return False

    def has_receive(self):
        """Implement EventLinkAbove.has_receive."""
        for top in self.top:
            try:
                if top.has_receive():
                    return True
            except AttributeError:
                pass
        return False

    def send(self, event):
        """Implement EventLinkAbove.send."""
        for top in self.top:
            try:
                top.send(event)
            except AttributeError:
                pass

    def read(self):
        """Implement StreamLinkAbove.read."""
        all_read = None
        for top in self.top:
            try:
                if all_read is None:
                    all_read = top.read()
                else:
                    all_read += top.read()
            except AttributeError:
                pass
        return all_read

    def write(self, event):
        """Implement StreamLinkAbove.write."""
        for top in self.top:
            try:
                top.write(event)
            except AttributeError:
                pass


class ManualPipe(Pipe):
    """Link to join two layers of EventLinks or StreamLinks together in a pipeline.

    This one requires manual synchronization between the links, by calling the
    sync method.
    """

    def __init__(self, *args):
        """Initialize members."""
        super().__init__(*args)
        self.next_clock_request = None
        self.last_clock_update = None

    # Synchronization

    def sync_bottom(self):
        """If any bottom links are also ManualPipes, sync them."""
        for link in self.bottom:
            try:
                link.sync()
            except AttributeError:  # link is not a ManualPipe
                pass

    def sync_top_links(self):
        """If any top links are also ManualPipes, sync them."""
        for link in self.top:
            try:
                link.sync()
            except AttributeError:  # link is not a ManualPipe
                pass

    def sync(self):
        """Synchronize the queues or buffers between the two layers.

        The receive() of bottom will be passed up to the to_receive() of top, and the
        to_send() of top will be passed down to the send() of bottom.

        Returns the earliest clock update requested by any processor within the
        pipe. This consists of the receivers in the links in the bottom layer
        and the senders in the links in the top layer.
        """
        self.sync_up()
        self.sync_down()
        return self.next_clock_request

    def sync_up(self):
        """Synchronize the queues or buffers from the bottom layer to the top layer.

        The receive() of bottom will be passed up to the to_receive() of top.

        Returns the earliest clock update requested by the receiver of any link
        in the bottom layer.
        """
        clock_requests = list(remove_none(self._sync_up(bottom) for bottom in self.bottom))
        if clock_requests:
            next_clock_request = min(clock_requests)
            self.next_clock_request = min(self.next_clock_request, next_clock_request)
        return self.next_clock_request

    def sync_down(self):
        """Synchronize the queues or buffers from the top layer to the bottom layer.

        The to_send() of top will be passed down to the send() of bottom.

        Returns the earliest clock update requested by the sender of any link
        in the top layer.
        """
        clock_requests = list(remove_none(self._sync_down(top) for top in self.top))
        if clock_requests:
            next_clock_request = min(clock_requests)
            self.next_clock_request = min(self.next_clock_request, next_clock_request)
        return self.next_clock_request

    def _sync_up(self, bottom):
        """Synchronize from a bottom link to the top layer.

        Returns the earliest clock update requested by the receiver of the bottom link.
        """
        earliest_clock_request = None
        self.sync_bottom()
        try:
            for event in bottom.receive_all():
                if isinstance(event, LinkClockRequest):
                    earliest_clock_request = min(earliest_clock_request, event)
                else:
                    self.receive_up(event)
        except AttributeError:
            pass
        try:
            self.read_up(bottom.read())
        except AttributeError:
            pass
        self.sync_top_links()
        return earliest_clock_request

    def _sync_down(self, top):
        """Synchronize from a top link to the bottom layer.

        Returns the earliest clock update requested by the sender of the top link.
        """
        earliest_clock_request = None
        self.sync_top_links()
        try:
            for event in top.to_send_all():
                if isinstance(event, LinkClockRequest):
                    earliest_clock_request = min(earliest_clock_request, event)
                else:
                    self.send_down(event)
        except AttributeError:
            pass
        try:
            self.write_down(top.to_write())
        except AttributeError:
            pass
        self.sync_bottom()
        return earliest_clock_request

    # Clocks

    def update_clock(self, time):
        """Update the clock of any ClockedLink and do any necessary processing."""
        if self.next_clock_request is not None and time >= self.next_clock_request:
            self.next_clock_request = None
        self.last_clock_update = time
        super().update_clock(time)
        return self.sync()


class AutomaticPipe(Pipe):
    """Link to join the queues of two EventLinks together in a pipeline.

    This one provides automatic synchronization between the links, by modifying
    the protocols.

    Warning: this works by monkey-patching the after_receive and after_send methods
    of EventLink, and the after_read and after_write methods of StreamLink! So
    if you want to use this pipe, don't put any custom logic in there.
    Warning: after you make the pipe, you shouldn't modify the bottom or top
    members; instead, make a new pipe to add new connections.
    """

    def __init__(self, bottom, top):
        """Initialize protocols.

        Instantiate this before sending any data to bottom or top, or else you will probably
        lose data!
        """
        super().__init__(bottom, top)
        self.connected_up = True
        self.connected_down = True
        for bottom in self.bottom:
            if hasattr(bottom, 'after_receive'):
                bottom.after_receive = self._after_receive
            if hasattr(bottom, 'after_read'):
                bottom.after_read = self._after_read
        for top in self.top:
            if hasattr(top, 'after_send'):
                top.after_send = self._after_send
            if hasattr(top, 'after_write'):
                top.after_write = self._after_write
        self.bottom_clocked = [
            link for link in self.bottom if hasattr(link, 'update_clock')
        ]
        self.top_clocked = [
            link for link in self.top if hasattr(link, 'update_clock')
        ]
        self.clocked = self.bottom_clocked or self.top_clocked
        self.next_clock_request = None
        self.last_clock_update = None

    def _after_receive(self, event):
        """Pass the processed event up to the top layer."""
        if isinstance(event, LinkClockRequest):
            self.next_clock_request = min(self.next_clock_request, event)
            # print(
            #     'AutomaticPipe updated next_clock_request to {}!'
            #     .format(self.next_clock_request)
            # )
            return
        if not self.connected_up:
            return
        # print('AutomaticPipe passing event up: {}'.format(event))
        self.receive_up(event)
        yield from proceed()  # proceed to process any additional events
        # print('AutomaticPipe after_receive finished yielding from proceed!')

    def _after_send(self, event):
        """Pass the processed event down to the bottom layer."""
        if isinstance(event, LinkClockRequest):
            self.next_clock_request = min(self.next_clock_request, event)
            # print(
            #     'AutomaticPipe updated next_clock_request to {}!'
            #     .format(self.next_clock_request)
            # )
            return
        if not self.connected_down:
            return
        # print('AutomaticPipe passing event down: {}'.format(event))
        self.send_down(event)
        yield from proceed()  # proceed to process any additional events
        # print('AutomaticPipe after_send finished yielding from proceed!')

    def _after_read(self, buffer):
        """Pass the processed buffer up to the top layer."""
        if not self.connected_up:
            return
        # print('AutomaticPipe passing buffer up: {}'.format(buffer))
        self.read_up(buffer)
        yield from wait()  # wait for more buffer activity

    def _after_write(self, buffer):
        """Pass the processed buffer down to the bottom layer."""
        if not self.connected_down:
            return
        # print('AutomaticPipe passing buffer down: {}'.format(buffer))
        self.write_down(buffer)
        yield from wait()  # wait for more buffer activity

    # Clocks

    def update_clock(self, time):
        """Update the clock of any ClockedLink and do any necessary processing."""
        self.last_clock_update = time
        self.update_clock_send(time)
        self.update_clock_receive(time)
        return self.next_clock_request

    def update_clock_send(self, time):
        """Update the clocks of any ClockedLink and do any necessary processing, for send.

        Updates the bottom clocks first, because automatic piping needs the bottom
        send processor to have the updated time before it automatically gets
        events from the top send processor.
        """
        if not self.clocked:
            return
        if self.next_clock_request is not None and time >= self.next_clock_request:
            self.next_clock_request = None
        for link in itertools.chain(self.bottom_clocked, self.top_clocked):
            link.send(LinkClockTime(time))
        return self.next_clock_request

    def update_clock_receive(self, time):
        """Update the clocks of any ClockedLink and do any necessary processing, for receive.

        Updates the top clocks first, because automatic piping needs the top
        receive processor to have the updated time before it automatically gets
        events from the bottom receive processor.
        """
        if not self.clocked:
            return
        if self.next_clock_request is not None and time >= self.next_clock_request:
            self.next_clock_request = None
        for link in itertools.chain(self.top_clocked, self.bottom_clocked):
            link.to_receive(LinkClockTime(time))
        return self.next_clock_request



"""Utilities for writing Phyllo protocol processors.

Processors can be formally described as nodes/processes of a Kahn process network
(see https://en.wikipedia.org/wiki/Kahn_process_networks for details), but
always with one input channel and one output channel.

Provides interfaces to write byte buffer-based stream protocol processors or
queue-based discrete event protocol processors.

The terminology here is a bit overloaded: protocol processors are called "protocols"
 n ohneio, but they are not network protocols. So instead we call them processors.
"""

# Builtins

import functools
import inspect
from collections import deque

# Packages

import ohneio


# Flow control


wait = ohneio.wait
_proceed = ohneio._Action('proceed')  # continue without waiting


def proceed():
    """Proceed without waiting for any action to be triggered on the consumer."""
    yield _proceed


# Event-based processors

class EventConsumer(ohneio.Consumer):
    """Generalized consumer which uses handles inputs and outputs on a deque.

    Allows the developer to send, read, and get the result of a processor function.

    This never needs to be instantiated, since this is internally done by the
    processor decorator.

    Adapted from ohneio.
    """

    def __init__(self, gen, input_factory=deque, output_factory=deque):
        """Initialize members."""
        self.gen = gen
        self.input = input_factory()
        self.output = output_factory()
        self.state = next(gen)
        if not isinstance(self.state, ohneio._Action):
            raise RuntimeError(
                "Can't yield anyting else than an action. Using `yield` instead "
                'of `yield from`?'
            )
        self.res = ohneio._no_result

    def _process(self):
        if self.has_result:
            return

        while self.state is ohneio._wait:
            self._next_state()
        while True:
            if self.state is ohneio._get_output:
                self._next_state(self.output)
            elif self.state is ohneio._get_input:
                self._next_state(self.input)
            elif self.state is _proceed:
                self._next_state()
            else:
                break

    def read(self):
        """Read and consume event from the output of the processor."""
        return self.output.popleft()

    def has_read(self):
        """Return events at the output of the processor, without consuming them."""
        return len(self.output) > 0

    def send(self, event):
        """Send event to the input of the processor."""
        self.input.append(event)
        self._process()

    def directly_to_read(self, event):
        """Add event to the output queue of the processor."""
        self.output.append(event)
        self._process()


def event_processor(func, input=deque, output=deque):
    """Wrap a Phyllo processor generator function as a decorator.

    Under the hood this wraps the generator inside an EventConsumer.
    """
    if not callable(func):
        raise ValueError('A processor needs to be a callable')
    if not inspect.isgeneratorfunction(func):
        raise ValueError('A processor needs to be a generator function')

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return EventConsumer(func(*args, **kwargs), input_factory=input, output_factory=output)

    return wrapper


def receive():
    """Receive and consume events.

    Receive and consume events from the processor input queue.
    """
    while True:
        input_queue = yield ohneio._get_input
        try:
            return input_queue.popleft()
        except IndexError:
            pass
        yield from wait()


def send(event):
    """Send event so that it can be consumed.

    Send event to the processor output queue.
    """
    output_queue = yield ohneio._get_output
    output_queue.append(event)


# ohneio-style stream-based processors

stream_processor = ohneio.protocol


def can_read(num_bytes=1):
    """Wait until we can receive and consume at least the specified number of bytes."""
    while True:
        input_buffer = yield ohneio._get_input
        if len(input_buffer) >= num_bytes:
            return
        else:
            pass
        yield from wait()


# We don't reimport ohneio.read because we provide an implementation whose
# interface is a superset of what is offered by ohneio.read.
def read(min_bytes=1, max_bytes=None):
    """Receive and consume the specified number of bytes.

    If min_bytes is nonzero, only returns when new bytes are on the input buffer.
    If max_bytes is None, then it reads everything available.
    """
    yield from can_read(min_bytes)  # never return an empty buffer
    if max_bytes is None:
        max_bytes = 0
    buffer = yield from ohneio.read(max_bytes)
    return buffer


def read_until(separator):
    """Generate a buffer of data until the next separator in the stream buffer.

    Waits for a complete frame to become available.
    """
    chunk_buffer = bytearray()
    while True:
        byte = yield from read(max_bytes=1)
        chunk_buffer.extend(byte)
        if chunk_buffer.endswith(separator):
            return bytes(chunk_buffer[:-len(separator)])


# We don't reimport ohneio.write because we provide an implementation which
# doesn't block the processor while waiting for the output to be entirely consumed.
def write(buffer):
    """Write data."""
    output = yield ohneio._get_output
    output.write(buffer)



"""Classes for building pipelines of layered links."""

# Builtins

# Packages

from phylline.links.links import GenericLinkAbove, GenericLinkBelow
from phylline.pipes import AutomaticPipe, ManualPipe
from phylline.processors import wait  # Any protocols
from phylline.util.interfaces import SetterProperty
from phylline.util.iterables import remove_none


# Pipelines


class Pipeline(GenericLinkBelow, GenericLinkAbove):
    """A linear pipeline which connects layers of links with Pipes."""

    def __init__(self, pipe_factory, *layers, name=None, **pipe_factory_kwargs):
        """Initialize the pipeline."""
        self.name = name
        self.layers = layers
        self.pipes = [
            pipe_factory(below, above, **pipe_factory_kwargs)
            for (below, above) in zip(self.layers[:-1], self.layers[1:])
        ]
        self.next_clock_request = None
        self.last_clock_update = None

    @property
    def bottom(self):
        """Return the bottom layer."""
        return self.layers[0]

    @property
    def top(self):
        """Return the bottom layer."""
        return self.layers[-1]

    def __repr__(self):
        """Return a string representation of the pipeline.

        Left-to-right order is from bottom to top.
        """
        layers = '|'.join('{}'.format(layer) for layer in self.layers)
        if self.name is not None:
            return '[{}: {}]'.format(self.name, layers)
        else:
            return '[{}]'.format(layers)

    # Clocks

    def update_clock(self, time):
        """Update the clock of any ClockedLink and do any necessary processing."""
        self.last_clock_update = time
        if self.clock_update_requested(time):
            self.next_clock_request = None
            # print('Reset next clock request to None!')
        for pipe in self.pipes:
            pipe.update_clock(time)

    def clock_update_requested(self, time):
        """Return whether a clock request has been requested by the given time."""
        return (
            self.next_clock_request is not None and time >= self.next_clock_request
        )

    # EventLink/StreamLink-like interface

    @SetterProperty
    def after_receive(self, handler):
        """Mimic EventLink.after_receive."""
        self.pipes[-1].after_receive = handler

    @SetterProperty
    def after_send(self, handler):
        """Mimic EventLink.after_receive."""
        self.pipes[0].after_send = handler

    @SetterProperty
    def after_read(self, handler):
        """Mimic StreamLink.after_read."""
        self.pipes[-1].after_read = handler

    @SetterProperty
    def after_write(self, handler):
        """Mimic StreamLink.after_write."""
        self.pipes[0].after_write = handler

    # Implement GenericLinkBelow

    def to_receive(self, event):
        """Implement EventLinkBelow.to_receive."""
        return self.pipes[0].to_receive(event)

    def to_send(self):
        """Implement EventLinkBelow.to_send."""
        return self.pipes[0].to_send()

    def has_to_send(self):
        """Implement EventLinkBelow.has_to_send."""
        return self.pipes[0].has_to_send()

    def to_read(self, event):
        """Implement StreamLinkBelow.to_read."""
        return self.pipes[0].to_read(event)

    def to_write(self):
        """Implement StreamLinkBelow.to_write."""
        return self.pipes[0].to_write()

    # Implement GenericLinkAbove

    def receive(self):
        """Implement EventLinkAbove.receive."""
        return self.pipes[-1].receive()

    def has_receive(self):
        """Implement EventLinkAbove.has_receive."""
        return self.pipes[-1].has_receive()

    def send(self, event):
        """Implement EventLinkAbove.send."""
        return self.pipes[-1].send(event)

    def read(self):
        """Implement StreamLinkAbove.read."""
        return self.pipes[-1].read()

    def write(self, event):
        """Implement StreamLinkAbove.write."""
        return self.pipes[-1].write(event)


class ManualPipeline(Pipeline):
    """A manually synchronized pipeline."""

    def __init__(self, *layers, pipe_factory=ManualPipe, **kwargs):
        """Initialize the pipeline."""
        super().__init__(pipe_factory, *layers, **kwargs)

    # Synchronization

    def sync(self):
        """Sync data from the lowest layer to the highest, then backwards.

        Returns the earliest clock update requested by any link in the pipeline.
        """
        # print('Sync up...')
        self.sync_up()
        # print('Sync down...')
        self.sync_down()
        return self.next_clock_request

    def sync_up(self):
        """Sync data from the lowest layer to the highest.

        Returns the earliest clock update requested by any bottom link in the pipeline.
        """
        clock_requests = list(remove_none(pipe.sync_up() for pipe in self.pipes))
        if clock_requests:
            next_clock_request = min(clock_requests)
            self.next_clock_request = min(self.next_clock_request, next_clock_request)
        return self.next_clock_request

    def sync_down(self):
        """Sync data from the highest layer to the lowest.

        Returns the earliest clock update requested by any top link in the pipeline.
        """
        clock_requests = list(remove_none(pipe.sync_down() for pipe in reversed(self.pipes)))
        if clock_requests:
            next_clock_request = min(clock_requests)
            self.next_clock_request = min(self.next_clock_request, next_clock_request)
        return self.next_clock_request

    # Clocks

    def update_clock(self, time):
        """Update the clock of any ClockedLink and do any necessary processing.

        Returns the earliest clock update requested by any processor within the pipeline.
        """
        super().update_clock(time)
        return self.sync()


class AutomaticPipeline(Pipeline):
    """A automatically synchronized pipeline."""

    def __init__(self, *layers, pipe_factory=AutomaticPipe, **kwargs):
        """Initialize the pipeline."""
        super().__init__(pipe_factory, *layers, **kwargs)
        self.pipes_clocked = [pipe for pipe in self.pipes if pipe.clocked]
        self.clocked = len(self.pipes_clocked) > 0

    # Clocks

    def update_clock(self, time):
        """Update the clock of any ClockedLink and do any necessary processing."""
        self.last_clock_update = time
        if not self.clocked:
            return
        if self.next_clock_request is not None and time >= self.next_clock_request:
            self.next_clock_request = None
            # print('Reset next clock request to None!')
        for pipe in self.pipes_clocked:
            next_clock_request = pipe.update_clock_send(time)
            if self.next_clock_request is None:
                self.next_clock_request = next_clock_request
            elif next_clock_request is not None:
                self.next_clock_request = min(self.next_clock_request, next_clock_request)
        for pipe in reversed(self.pipes_clocked):
            next_clock_request = pipe.update_clock_receive(time)
            if self.next_clock_request is None:
                self.next_clock_request = next_clock_request
            elif next_clock_request is not None:
                self.next_clock_request = min(self.next_clock_request, next_clock_request)
        return self.next_clock_request


class PipelineBottomCoupler(object):
    """A pipeline-to-pipeline coupler which connects two pipelines by their bottoms."""

    def __init__(self, pipeline_one, pipeline_two):
        """Initialize members."""
        self.pipeline_one = pipeline_one
        self.pipeline_two = pipeline_two
        self.is_manual_one = isinstance(self.pipeline_one, ManualPipeline)
        self.is_manual_two = isinstance(self.pipeline_two, ManualPipeline)
        if isinstance(self.pipeline_one, AutomaticPipeline):
            self.pipeline_one.after_write = self._write_one
        if isinstance(self.pipeline_two, AutomaticPipeline):
            self.pipeline_two.after_write = self._write_two

    def __repr__(self):
        """Represent the coupler as a string."""
        return '╔{}\n╚{}'.format(self.pipeline_one, self.pipeline_two)

    def update_clock(self, clock_time):
        """Update the pipeline's clock.

        If the pipeline is a manual pipeline, it will also sync the pipeline and write
        any data at the bottom to the connection.
        """
        self.pipeline_one.update_clock(clock_time)
        self.pipeline_two.update_clock(clock_time)
        if self.is_manual_one:
            self.write_one()
        if self.is_manual_two:
            self.write_two()

    def _write_one(self, buffer):
        """Write the buffer to the connection.

        This is used to overwrite the after_write of the bottom of the pipeline
        if it's an automatic pipeline.
        """
        # print('Sending to pipeline two: {}'.format(hex_bytes(buffer)))
        self.pipeline_two.to_read(buffer)
        yield from wait()

    def _write_two(self, buffer):
        """Write the buffer to the connection.

        This is used to overwrite the after_write of the bottom of the pipeline
        if it's an automatic pipeline.
        """
        # print('Sending to pipeline one: {}'.format(hex_bytes(buffer)))
        self.pipeline_one.to_read(buffer)
        yield from wait()

    def write_one(self):
        """Update the write side of the coupler.

        This is used for manual writing from the bottom of the pipeline if it's not
        an automatic pipeline.
        """
        data = self.pipeline_one.to_write()
        if data:
            self.pipeline_two.to_read(data)
        return data

    def write_two(self):
        """Update the write side of the coupler.

        This is used for manual writing from the bottom of the pipeline if it's not
        an automatic pipeline.
        """
        data = self.pipeline_two.to_write()
        if data:
            self.pipeline_one.to_read(data)
        return data



"""Base classes and utilities for writing communication links.

Links consist of a pair of processors: one processor to process incoming data from
the layer below and surface any resulting data to the layer above, and another to
process incoming data from the layer above and surface any resulting data to the
layer below. Then a link is a pair of independent Kahn process network nodes
which process data in opposite directions.

Provides interfaces to write byte buffer-based stream links or queue-based
discrete event links.
"""

# Builtins

from abc import abstractmethod

# Packages

from phylline.processors import event_processor, receive, send
from phylline.util.logging import hex_bytes


# Events

class LinkEvent(object):
    """Base class to represent any event an event link might produce or consume.

    context contains metadata as specified by the link.
    instance should be a reference to the link instance which produced the event.
    previous should be a reference to the previous LinkEvent or data which caused
        this event to be created.
    """

    def __init__(self, context=None, instance=None, previous=None):
        """Initialize members."""
        if context is None:
            context = {}
        self.context = context
        self.link = str(instance)
        self.previous = previous

    def __str__(self):
        """Represent as a string."""
        return '{} from ({}) {}{}'.format(
            self.__class__.__qualname__, self.link,
            'with context({})'.format(self.context) if self.context else '',
            ', due to ({})'.format(self.previous) if self.previous is not None else ''
        )

    def __repr__(self):
        """Represent as a string."""
        return '{}(context={}, instance={}, previous={})'.format(
            self.__class__.__qualname__, self.context, self.link, self.previous
        )


class LinkData(LinkEvent):
    """Event indicating a unit of data which was passed up or down a layer.

    data should be a reference to the data being passed.
    direction should be a string indicating the direction in which the data is
        being passed.
    """

    def __init__(
        self, data, direction='up', context=None, instance=None, previous=None
    ):
        """Initialize members."""
        super().__init__(context=context, instance=instance, previous=previous)
        self.data = data
        self.direction = direction

    def __str__(self):
        """Represent as a string."""
        return '{} passed {} in ({}){}: {}{}'.format(
            self.__class__.__qualname__,
            self.direction, self.link,
            ' with context({})'.format(self.context) if self.context else '',
            hex_bytes(self.data) if isinstance(self.data, (bytes, bytearray)) else self.data,
            ', due to ({})'.format(self.previous) if self.previous is not None else ''
        )

    def __repr__(self):
        """Represent as a string."""
        return '{}({}, direction={}, context={}, instance={}, previous={})'.format(
            self.__class__.__qualname__, self.data, self.direction, self.context,
            self.link, self.previous
        )


class LinkException(LinkEvent):
    """Event indicating an exception passed up or down a layer.

    exception should be a reference to the exception being passed.
    direction should be a string indicating the direction in which the data is
        being passed.
    """

    def __init__(
        self, exception, direction='up', context=None, instance=None, previous=None
    ):
        """Initialize members."""
        super().__init__(context=context, instance=instance, previous=previous)
        self.exception = exception
        self.direction = direction

    def __str__(self):
        """Represent as a string."""
        return '{} passed {} in ({}){}: {}{}'.format(
            self.__class__.__qualname__,
            self.direction, self.link,
            ' with context({})'.format(self.context) if self.context else '',
            self.exception,
            ', due to ({})'.format(self.previous) if self.previous is not None else ''
        )

    def __repr__(self):
        """Represent as a string."""
        return '{}({}, direction={}, context={}, instance={}, previous={})'.format(
            self.__class__.__qualname__, self.exception, self.direction, self.context,
            self.link, self.previous
        )


# Link interfaces


class EventLinkAbove(object):
    """Interface for exposing an EventLink-like interface for the layer above.

    This interface is analogous to the interface for conn1 of a duplex
    multiprocessing.Pipe.
    """

    @abstractmethod
    def receive(self):
        """Return the next received event, while consuming them."""
        pass

    def receive_all(self):
        """Generate all received events, while consuming them."""
        while self.has_receive():
            yield self.receive()

    @abstractmethod
    def has_receive(self):
        """Return whether received events are available, without consuming them."""
        pass

    @abstractmethod
    def send(self, event):
        """Send event on the link."""
        pass


class EventLinkBelow(object):
    """Interface for exposing an EventLink-like interface for the layer below.

    This interface is analogous to the interface for conn2 of a duplex
    multiprocessing.Pipe.
    """

    @abstractmethod
    def to_receive(self, event):
        """Receive an event on the link."""
        pass

    @abstractmethod
    def to_send(self):
        """Return the next event to send, while consuming it."""
        pass

    def to_send_all(self):
        """Generate all events to send, while consuming them."""
        while self.has_to_send():
            yield self.to_send()

    @abstractmethod
    def has_to_send(self):
        """Return available events to send, without consuming them."""
        pass


# Link base classes


class DataEventLink(object):
    """Support for exposing an EventLink-like interface supporting LinkData events."""

    def send_data(self, data, direction='down', context=None):
        """Send data from above as a LinkData event."""
        self._sender.send(LinkData(
            data, direction=direction, context=context, instance=self
        ))

    def get_link_data(self, event, direction, context=None):
        """Return a LinkData event from the provided event.

        The provided event must be either a LinkData event or some kind of data
        which is not a LinkEvent.
        """
        if isinstance(event, LinkData):
            return LinkData(
                event.data, direction=direction,
                context=event.context.copy() if context is None else context,
                instance=self, previous=event
            )
        elif not isinstance(event, LinkEvent):
            return LinkData(event, direction=direction, context=context, instance=self)

    def make_link_data(self, data, direction, previous, context=None):
        """Return a LinkData event from the provided data."""
        return LinkData(
            data, direction=direction, context=context, instance=self,
            previous=previous
        )

    def make_link_exception(self, exception, direction, previous, context=None):
        """Return a LinkException event."""
        return LinkException(
            exception, direction=direction, context=context, instance=self,
            previous=previous
        )


class EventLink(EventLinkBelow, EventLinkAbove, DataEventLink):
    """A duplex link for sending and receiving events.

    EventLinks have a pair of event queues: one for transforming data from
    the layer below the link for consumption by the layer above the link, and
    one for transforming data from the layer above the link for transmission by
    the layer below the link.
    When the layer below the link has an event ready for consumption by the
    EventLink, it calls the to_receive method with this event. The EventLink then
    processes this event, and when it is complete, the layer above the link can
    process any events surfaced from this processing by calling the receive method
    as an iterable.
    Vice versa, the layer above the link can provide an event for EventLink to
    process for sending (by the layer below the link) by calling its send method,
    and the layer below the link can call the to_send method as an iterable to
    get any events it needs to process for sending.

    The default behavior is to pass the event through in both directions, but
    data is wrapped in a DataEvent. To pass the event through without any
    transformation, enable the receiver_event_passthrough and
    sender_event_passthrough flags.  To change this behavior further, either
    override the class with a custom sender and receiver processor, or provide a
    custom processor upon construction.
    """

    def __init__(
        self, name=None, receiver_processor=None, sender_processor=None,
        receiver_processor_args=(), receiver_processor_kwargs={},
        sender_processor_args=(), sender_processor_kwargs={},
        receiver_event_passthrough=False, sender_event_passthrough=False
    ):
        """Initialize reader and writer processors."""
        super().__init__()
        self.name = name
        if receiver_processor is None:
            receiver_processor = self.receiver_processor
        self._receiver = receiver_processor(
            *receiver_processor_args, **receiver_processor_kwargs
        )
        if sender_processor is None:
            sender_processor = self.sender_processor
        self._sender = sender_processor(
            *sender_processor_args, **sender_processor_kwargs
        )
        self.receiver_event_passthrough = receiver_event_passthrough
        self.sender_event_passthrough = sender_event_passthrough

    @property
    def _using_own_receiver_processor(self):
        return (
            self.receiver_processor.__qualname__
            == '{}.receiver_processor'.format(self.__class__.__qualname__)
        )

    @property
    def _using_own_sender_processor(self):
        return (
            self.sender_processor.__qualname__
            == '{}.sender_processor'.format(self.__class__.__qualname__)
        )

    def __repr__(self):
        """Return a string representation of the link."""
        if self.name is not None:
            return '⇌□ {}({}) □⇌'.format(self.__class__.__qualname__, self.name)
        else:
            return '⇌□ {} □⇌'.format(self.__class__.__qualname__)

    # Implement EventLinkAbove

    def receive(self):
        """Implement EventLinkAbove.receive."""
        return self._receiver.read()

    def has_receive(self):
        """Implement EventLinkAbove.has_receive."""
        return self._receiver.has_read()

    def send(self, event):
        """Implement EventLinkAbove.send."""
        self._sender.send(event)

    # Implement EventLinkBelow

    def to_receive(self, event):
        """Implement EventLinkBelow.to_receive."""
        self._receiver.send(event)

    def to_send(self):
        """Implement EventLinkBelow.to_send."""
        return self._sender.read()

    def has_to_send(self):
        """Implement EventLinkBelow.has_to_send."""
        return self._sender.has_read()

    # Utilities for writing receive and send processors

    def directly_receive(self, event):
        """Add an event directly to the receive queue.

        Intended for use by the sender processor to short-circuit the receiver
        processor and directly add something to the receive queue.
        """
        self._receiver.directly_to_read(event)

    def directly_to_send(self, event):
        """Add an event directly to the receive queue.

        Intended for use by the receiver processor to short-circuit the sender
        processor and directly add something to the to_send queue.
        """
        self._sender.directly_to_read(event)

    def after_receive(self, event):
        """Enqueue the event for the layer above for consumption by that layer.

        This gets overridden to change the receiver's behavior to send the event.
        Note that this gets monkey-patched by other links and link utilities
        which combine links, such as ChunkedStreamLink and AutomaticPipe!
        """
        # print('Event link after_receive: {}'.format(event))
        yield from send(event)

    def after_send(self, event):
        """Enqueue the event for the layer below for consumption by that layer.

        This gets overridden to change the receiver's behavior to send the event.
        Note that this gets monkey-patched by other links and link utilities
        which combine links, such as ChunkedStreamLink and AutomaticPipe!
        """
        # print('Event link after_send: {}'.format(event))
        yield from send(event)

    # Receive and send processors

    @event_processor
    def receiver_processor(self):
        """Event receiver processor.

        Make sure to yield from after_receive() with any generated event to expose
        it for for consumption by the layer above.
        """
        while True:
            event = yield from receive()
            if self.receiver_event_passthrough or isinstance(event, LinkException):
                # print('Event Link received: {}'.format(event))
                yield from self.after_receive(event)
            else:
                data_event = self.get_link_data(event, 'up')
                # print('Event Link received: {}'.format(data_event))
                data_event = self.make_link_data(data_event.data, 'up', event)
                yield from self.after_receive(data_event)

    @event_processor
    def sender_processor(self):
        """Event sender processor.

        Make sure to yield from after_send() with any generated event to expose
        it for consumption by the layer below.
        """
        while True:
            event = yield from receive()
            if self.sender_event_passthrough:
                # print('Event Link sending: {}'.format(event))
                yield from self.after_send(event)
            else:
                data_event = self.get_link_data(event, 'down')
                # print('Event Link sending: {}'.format(data_event))
                data_event = self.make_link_data(data_event.data, 'down', event)
                yield from self.after_send(data_event)



"""Test the util.logging module."""

# Builtins

# Packages

from phylline.util.logging import hex_bytes

import pytest


def test_hex_bytes_negative_truncate():
    """Test whether the hex_bytes correctly handles invalid truncate parameter values."""
    with pytest.raises(ValueError):
        hex_bytes(bytes([0x00, 0x01, 0x02]), truncate=-1)
    with pytest.raises(ValueError):
        hex_bytes(bytes([]), truncate=-1)


def test_hex_bytes_empty():
    """Test whether the hex_bytes correctly renders empty buffers."""
    assert hex_bytes(bytes([]), truncate=1) == '(0 bytes)'
    assert hex_bytes(bytes([]), truncate=0) == '(0 bytes)'


def test_hex_bytes_untruncated():
    """Test whether the hex_bytes correctly renders buffers without truncation."""
    short = bytes([0x00, 0x01, 0x02])
    assert(hex_bytes(short, truncate=None)) == '0x00 0x01 0x02'
    buffer = bytes(range(5))
    assert hex_bytes(buffer, truncate=6) == '0x00 0x01 0x02 0x03 0x04'
    assert hex_bytes(buffer, truncate=5) == '0x00 0x01 0x02 0x03 0x04'


def test_hex_bytes_truncated():
    """Test whether the hex_bytes correctly renders buffers with truncation."""
    buffer = bytes(range(5))
    assert hex_bytes(buffer, truncate=4) == '0x00 0x01 0x02 0x03 ...(1 more byte)'
    assert hex_bytes(buffer, truncate=3) == '0x00 0x01 0x02 ...(2 more bytes)'
    assert hex_bytes(buffer, truncate=2) == '0x00 0x01 ...(3 more bytes)'
    assert hex_bytes(buffer, truncate=1) == '0x00 ...(4 more bytes)'
    assert hex_bytes(buffer, truncate=0) == '(5 bytes)'
    assert hex_bytes(bytes([0x00]), truncate=0) == '(1 byte)'



"""Base classes and utilities for writing communication links.

Links consist of a pair of processors: one processor to process incoming data from
the layer below and surface any resulting data to the layer above, and another to
process incoming data from the layer above and surface any resulting data to the
layer below. Then a link is a pair of independent Kahn process network nodes
which process data in opposite directions.

Provides interfaces to write byte buffer-based stream links or queue-based
discrete event links.
"""

# Builtins

import collections
import math

# Packages

from phylline.links.events import EventLink, LinkEvent, LinkException
from phylline.processors import event_processor, receive
from phylline.util.timing import Clock, TimeoutTimer


# Events


class LinkClockTime(LinkEvent):
    """Event indicating the current time to advance the event link's clock to.

    clock_time should be the current time to advance the clock to.
    """

    def __init__(self, clock_time, context=None, instance=None, previous=None):
        """Initialize members."""
        super().__init__(context=context, instance=instance, previous=previous)
        self.clock_time = clock_time

    def __str__(self):
        """Represent as a string."""
        return '{} for ({}){}: {}{}'.format(
            self.__class__.__qualname__, self.link,
            ' with context({})'.format(self.context) if self.context else '',
            self.clock_time,
            ', due to ({})'.format(self.previous) if self.previous is not None else ''
        )

    def __repr__(self):
        """Represent as a string."""
        return '{}({}, context={}, instance={}, previous={})'.format(
            self.__class__.__qualname__, self.clock_time, self.context, self.link, self.previous
        )


class LinkClockRequest(LinkEvent):
    """Event indicating the next time by which the event link needs a clock update."""

    def __init__(self, requested_time, context=None, instance=None, previous=None):
        """Initialize members."""
        super().__init__(context=context, instance=instance, previous=previous)
        self.requested_time = requested_time

    def __str__(self):
        """Represent as a string."""
        return '{} for ({}){}: {}{}'.format(
            self.__class__.__qualname__, self.link,
            ' with context({})'.format(self.context) if self.context else '',
            self.requested_time,
            ', due to ({})'.format(self.previous) if self.previous is not None else ''
        )

    def __repr__(self):
        """Represent as a string."""
        return '{}({}, context={}, instance={}, previous={})'.format(
            self.__class__.__qualname__, self.requested_time, self.context, self.link, self.previous
        )

    # Comparison operators

    def __eq__(self, other):
        """Check equality by the value of the requested time.

        Any clock request is not equal to None.
        """
        if other is None:
            return False
        if isinstance(other, LinkClockRequest):
            return self.requested_time == other.requested_time
        else:
            return self.requested_time == other

    def __ne__(self, other):
        """Check not-equality by the value of the requested time."""
        return not (self == other)

    def __lt__(self, other):
        """Check less-than by the value of the requested time.

        Any clock request is less than None.
        """
        if other is None:
            return True
        if isinstance(other, LinkClockRequest):
            return self.requested_time < other.requested_time
        else:
            return self.requested_time < other

    def __gt__(self, other):
        """Check greater-than by the value of the requested time.

        No clock request is greater than None.
        """
        if other is None:
            return False
        if isinstance(other, LinkClockRequest):
            return self.requested_time > other.requested_time
        else:
            return self.requested_time > other

    def __le__(self, other):
        """Check less-than-or-equal-to by the value of the requested time."""
        return self < other or self == other

    def __ge__(self, other):
        """Check greater-than-or-equal-to by the value of the requested time."""
        return self > other or self == other


# Clocked Links


class ClockedLink(object):
    """Support for implementing an EventLink which runs on an internal clock."""

    def __init__(self, *args, clock_start=0.0, **kwargs):
        """Initialize members."""
        super().__init__(*args, **kwargs)
        self.clock = Clock(time=clock_start)
        self.next_clock_request = TimeoutTimer(clock=self.clock)

    # Public interface

    def update_clock(self, time):
        """Update the clock of the link and do any necessary processing."""
        self.to_receive(LinkClockTime(time, instance=self))
        self.send(LinkClockTime(time, instance=self))

    # Internal methods for implementers

    def make_timer(self, delay):
        """Make a timer on the processor clock."""
        return TimeoutTimer(timeout=delay, clock=self.clock)

    def get_clock_time(self, event):
        """Extract any provided clock time from the event."""
        if isinstance(event, LinkClockTime):
            return event.clock_time
        return None  # No clock time provided

    def update_clock_time(self, event):
        """Return the new clock time, updated with any clock time in the event."""
        clock_time = self.get_clock_time(event)
        if clock_time is None:
            return
        # print('Updating clock to: {}'.format(clock_time))
        self.clock.update(clock_time)
        if self.next_clock_request.timed_out:
            self.next_clock_request.reset_and_stop()

    def make_clock_request(self, time, context={}, previous=None):
        """Return a LinkClockRequest if time is different from the last request."""
        if (
            self.next_clock_request.running
            and math.isclose(self.next_clock_request.timeout_time, time)
        ):
            return None
        self.next_clock_request.start(timeout=time - self.clock.time)
        return LinkClockRequest(
            time, context={'time': self.clock.time, **context}, instance=self,
            previous=previous
        )


class EventDelayer(object):
    """A helper class which passes events through after a time delay."""

    def __init__(self, processor, clock, delay):
        """Initialize members."""
        self.processor = processor
        self.clock = clock
        self.delay = delay
        self.in_flight = collections.deque()

    def make_timer(self):
        """Make a timer on the clock."""
        return TimeoutTimer(timeout=self.delay, clock=self.clock)

    def enqueue_event(self, event):
        """Add event to the in-flight queue of delayed events."""
        event.context['to_{}_time'.format(self.processor)] = self.clock.time
        timer = self.make_timer()
        timer.start()
        event.context['intended_{}_time'.format(self.processor)] = timer.timeout_time
        # print('EventDelayer for {} to receive at {}: {}'.format(
        #     self.processor, timer.timeout_time, event
        # ))
        self.in_flight.append({'event': event, 'timer': timer})

    def flush_events(self):
        """Dequeue and yield all events which have satisfied their delays."""
        # print(
        #     'In-flight send events at {}: {}'
        #     .format(self.clock.time, self.in_flight)
        # )
        while self.in_flight and self.in_flight[0]['timer'].timed_out:
            event = self.in_flight.popleft()['event']
            event.context['actual_{}_time'.format(self.processor)] = self.clock.time
            # print(
            #     'EventDelayer {}s at {}: {}'
            #     .format(self.processor, self.clock.time, event)
            # )
            yield event
        # print(
        #     'Post-flush in-flight send events at {}: {}'
        #     .format(self.clock.time, self.in_flight)
        # )

    @property
    def next_in_flight_event(self):
        """Return the next in-flight event to send after the appropriate delay."""
        return self.in_flight[0]


class DelayedEventLink(ClockedLink, EventLink):
    """An EventLink which passes events through only after a time delay.

    The time delay for the receive depends on regularly receiving timing events
    from to_receive. The time delay for the send depends on regularly receiving
    timing events from send.
    """

    def __init__(self, clock_start=0.0, receive_delay=1.0, send_delay=1.0):
        """Initialize members."""
        super().__init__(clock_start=clock_start)
        self._delayers = {
            'up': EventDelayer('receive', self.clock, receive_delay),
            'down': EventDelayer('send', self.clock, send_delay)
        }

    # Receive and send processors

    def _enqueue_data_event(self, event, direction):
        if isinstance(event, LinkException):
            return
        data_event = self.get_link_data(event, direction)
        if data_event is None:
            return

        data_event = self.make_link_data(data_event.data, direction, event)
        self._delayers[direction].enqueue_event(data_event)

    def _flush_event_queue(self, direction):
        delayer = self._delayers[direction]
        # print('Flushing {}...'.format(delayer.processor))
        for data_event in delayer.flush_events():
            yield from getattr(self, 'after_{}'.format(delayer.processor))(data_event)
            # print('Finished yielding from after_{}!'.format(processor))
        # print('Done flushing {}!'.format(processor))

    def _issue_clock_request(self, direction):
        delayer = self._delayers[direction]
        if not delayer.in_flight:
            return

        next_in_flight = delayer.next_in_flight_event
        clock_request = self.make_clock_request(
            next_in_flight['timer'].timeout_time,
            previous=next_in_flight['event'].previous
        )
        if clock_request is not None:
            yield from getattr(self, 'after_{}'.format(delayer.processor))(clock_request)

    @event_processor
    def receiver_processor(self):
        """Event receiver processor.

        Make sure to yield from after_receive() at the end of the loop to expose
        any generated events for consumption by the layer above.
        """
        while True:
            event = yield from receive()
            self.update_clock_time(event)
            self._enqueue_data_event(event, 'up')
            yield from self._flush_event_queue('up')
            yield from self._issue_clock_request('up')

    @event_processor
    def sender_processor(self):
        """Event sender processor.

        Make sure to yield from after_send() at the end of the loop to expose
        any generated events for consumption by the layer below.
        """
        while True:
            event = yield from receive()
            self.update_clock_time(event)
            self._enqueue_data_event(event, 'down')
            yield from self._flush_event_queue('down')
            yield from self._issue_clock_request('down')



"""Test the processors module."""

# Builtins

# Packages

from phylline.processors import event_processor, receive, send
from phylline.processors import proceed, wait
from phylline.processors import read, read_until, stream_processor, write

import pytest


# Basic event processor


@event_processor
def incrementer(last_received=[None]):
    """Increment every received number."""
    while True:
        number = yield from receive()
        yield from send(number + 1)
        last_received[0] = number


@event_processor
def incrementer_proceed(last_received=[None]):
    """Increment every received number."""
    while True:
        number = yield from receive()
        yield from send(number + 1)
        yield from proceed()  # this basically does nothing
        last_received[0] = number


def assert_incrementing(processor, last_received, number):
    """Test whether the processor increments the input."""
    processor.send(number)
    assert last_received == [number]
    assert processor.has_read()
    assert processor.read() == number + 1
    assert not processor.has_read()


@pytest.mark.parametrize('incrementer_processor', [
    incrementer,
    incrementer_proceed
])
def test_event_processor(incrementer_processor):
    """Test whether the event_processor works correctly with sends and receives."""
    last_received = [None]
    processor = incrementer_processor(last_received=last_received)

    # Individual processing
    assert_incrementing(processor, last_received, 1)
    assert_incrementing(processor, last_received, -2)

    # Batch processing
    for i in range(5):
        processor.send(i)
        assert last_received == [i]
    for i in range(5):
        assert processor.has_read()
        assert processor.read() == i + 1
    assert not processor.has_read()

    processor.directly_to_read(5)  # bypass the incrementer and add to read queue
    assert processor.has_read()
    assert processor.read() == 5


# Event processor with waiting


@event_processor
def incrementer_wait(last_received=[None]):
    """Increment every received number."""
    while True:
        number = yield from receive()
        yield from send(number + 1)
        yield from wait()  # wait until data is added to the receive queue before proceeding
        last_received[0] = number


def test_event_processor_wait():
    """Test whether the event_processor works correctly with sends and receives."""
    last_received = [None]
    processor = incrementer_wait(last_received=last_received)

    # Individual processing
    processor.send(1)
    assert last_received == [None]
    assert processor.read() == 2
    processor.send(2)
    assert last_received == [1]
    assert processor.read() == 3


# Event processor with return value


@event_processor
def incrementer_return():
    """Increment every received number."""
    number = yield from receive()
    yield from send(number + 1)
    return 10 * number


def test_event_processor_return():
    """Test whether the event_processor works correctly with returns."""
    processor = incrementer_return()

    processor.send(1)
    assert processor.read() == 2
    assert processor.get_result() == 10


# Invalid event processors


def test_event_processor_uncallable():
    """Test whether the event_processor decorator handles invalid inputs correctly."""
    with pytest.raises(ValueError):
        event_processor(5)


def test_event_processor_nongenerator():
    """Test whether the event_processor decorator handles invalid functions correctly."""
    with pytest.raises(ValueError):
        @event_processor
        def invalid_nongenerator():
            return 2


def test_event_processor_invalid_yield():
    """Test whether the event_processor decorator handles invalid yields correctly."""
    @event_processor
    def invalid_early():
        yield 42
    with pytest.raises(RuntimeError):
        processor = invalid_early()

    @event_processor
    def invalid_late():
        while True:
            number = yield from receive()
            yield number + 1
    processor = invalid_late()
    with pytest.raises(RuntimeError):
        processor.send(1)


# Basic stream processor


@stream_processor
def chunker():
    """Split the stream into chunks."""
    while True:
        buffer = yield from read_until(b'\0')
        yield from write(buffer)


@stream_processor
def passthrough():
    """Split the stream into chunks."""
    while True:
        buffer = yield from read()
        yield from write(buffer)


def test_stream_processor():
    """Test whether the event_processor works correctly with sends and receives."""
    processor = chunker()

    processor.send(b'\1\2\3\4')
    assert processor.read() == b''
    processor.send(b'\0')
    assert processor.read() == b'\1\2\3\4'
    processor.send(b'\1\2\0\3\4')
    assert processor.read() == b'\1\2'
    processor.send(b'\0\1\2\0\3\4\0')
    assert processor.read() == b'\3\4\1\2\3\4'

    processor = passthrough()
    processor.send(b'\1\2\3\4')
    assert processor.read() == b'\1\2\3\4'



"""Base classes and utilities for writing communication links.

Links consist of a pair of processors: one processor to process incoming data from
the layer below and surface any resulting data to the layer above, and another to
process incoming data from the layer above and surface any resulting data to the
layer below. Then a link is a pair of independent Kahn process network nodes
which process data in opposite directions.

Provides interfaces to write byte buffer-based stream links or queue-based
discrete event links.
"""

# Builtins

# Packages

from phylline.links.events import DataEventLink, EventLink, LinkException
from phylline.links.links import GenericLinkAbove, GenericLinkBelow
from phylline.links.streams import StreamLink
from phylline.processors import event_processor, receive
from phylline.processors import read, stream_processor
from phylline.processors import wait


class TopLoopbackLink(GenericLinkBelow, DataEventLink):
    """A data sink to echo events and stream buffers received from below to send back down.

    Note that the events received by to_receive are only echoed back on to_send,
    and buffers read by to_read are only echoed back on to_write.
    """

    def __init__(self, name=None):
        """Initialize members."""
        self.name = name
        self._event_link = EventLink(receiver_processor=self.receiver_processor)
        self._stream_link = StreamLink(reader_processor=self.reader_processor)

    def __repr__(self):
        """Return a string representation of the link."""
        if self.name is not None:
            return '⇌* {}({})'.format(self.__class__.__qualname__, self.name)
        else:
            return '⇌* {}'.format(self.__class__.__qualname__)

    # Implement EventLinkBelow

    def to_receive(self, event):
        """Implement EventLinkBelow."""
        self._event_link.to_receive(event)

    def to_send(self):
        """Implement EventLinkBelow."""
        return self._event_link.to_send()

    def has_to_send(self):
        """Implement EventLinkBelow."""
        return self._event_link.has_to_send()

    # Implement StreamLinkBelow

    def to_read(self, bytes_data):
        """Implement StreamLinkBelow."""
        self._stream_link.to_read(bytes_data)

    def to_write(self):
        """Implement StreamLinkBelow."""
        return self._stream_link.to_write()

    # Read and write processors

    def after_send(self, event):
        """Do nothing."""
        yield from wait()

    def after_write(self, buffer):
        """Do nothing."""
        yield from wait()

    @event_processor
    def receiver_processor(self):
        """Event receiver echo processor."""
        while True:
            event = yield from receive()
            if isinstance(event, LinkException):
                print('Ignoring exception: {}'.format(event))
                continue
            data_event = self.get_link_data(event, 'loopback')
            # print('Echoing: {}'.format(data_event))
            data_event = self.make_link_data(data_event.data, 'loopback', event)
            self._event_link._sender.send(data_event)
            yield from self.after_send(event)  # for correctness in automatic pipeline

    @stream_processor
    def reader_processor(self):
        """Stream reader echo processor."""
        while True:
            buffer = yield from read()
            # print('Echoing: {}'.format(buffer))
            self._stream_link._writer.send(buffer)
            yield from self.after_write(buffer)  # for correctness in automatic pipeline


class BottomLoopbackLink(GenericLinkAbove, DataEventLink):
    """A data sink to echo events and stream buffers sent from above to receive back up.

    Note that the events sent by send are only echoed back on receive,
    and buffers read by write are only echoed back on read.
    """

    def __init__(self, name=None):
        """Initialize members."""
        self.name = name
        self._event_link = EventLink(sender_processor=self.sender_processor)
        self._stream_link = StreamLink(writer_processor=self.writer_processor)

    def __repr__(self):
        """Return a string representation of the link."""
        if self.name is not None:
            return '{}({}) *⇌'.format(self.__class__.__qualname__, self.name)
        else:
            return '{} *⇌'.format(self.__class__.__qualname__)

    # Implement EventLinkAbove

    def receive(self):
        """Implement EventLinkAbove.receive."""
        return self._event_link.receive()

    def has_receive(self):
        """Implement EventLinkAbove.has_receive."""
        return self._event_link.has_receive()

    def send(self, event):
        """Implement EventLinkAbove.send."""
        self._event_link.send(event)

    # Implement StreamLinkAbove

    def read(self):
        """Implement StreamLinkAbove.read."""
        return self._stream_link.read()

    def write(self, bytes_data):
        """Implement StreamLinkAbove.write."""
        self._stream_link.write(bytes_data)

    # Read and write processors

    def after_receive(self, event):
        """Do nothing."""
        yield from wait()

    def after_read(self, buffer):
        """Do nothing."""
        yield from wait()

    @event_processor
    def sender_processor(self):
        """Event sender echo processor."""
        while True:
            event = yield from receive()
            data_event = self.get_link_data(event, 'loopback')
            data_event = self.make_link_data(data_event.data, 'loopback', event)
            self._event_link._receiver.send(data_event)
            yield from self.after_receive(data_event)  # for correctness in automatic pipeline

    @stream_processor
    def writer_processor(self):
        """Stream writer echo processor."""
        while True:
            buffer = yield from read()
            self._stream_link._reader.send(buffer)
            yield from self.after_read(buffer)  # needed for correctness in automatic pipeline



"""Test the util.iterables module."""

# Builtins

# Packages

from phylline.util.iterables import make_collection, remove_none


def test_make_collection_singleton():
    """Test whether the make_collection function makes collections from singletons."""
    assert make_collection(42) != 42
    assert make_collection(42) == [42]
    assert make_collection(42) != (42,)

    assert make_collection(42, type=tuple) != 42
    assert make_collection(42, type=tuple) != [42]
    assert make_collection(42, type=tuple) == (42,)


def test_make_collection_iterable():
    """Test whether the make_collection function makes collections from iterables."""
    assert make_collection(range(5)) != range(5)
    assert make_collection(range(5)) == list(range(5))
    assert make_collection(range(5)) != tuple(range(5))

    assert make_collection(range(5), type=tuple) != range(5)
    assert make_collection(range(5), type=tuple) != list(range(5))
    assert make_collection(range(5), type=tuple) == tuple(range(5))


def test_remove_none():
    """Test whether remove_none removes Nones correctly."""
    assert len(tuple(range(5))) == len(tuple(remove_none(range(5))))
    for (initial, filtered) in zip(range(5), remove_none(range(5))):
        assert initial == filtered
    assert len(tuple(remove_none([1, 2, None, 3]))) == 3
    assert tuple(remove_none([1, 2, None, 3])) == (1, 2, 3)



"""Test the util.interfaces module."""

# Builtins

import abc

# Packages

from phylline.util.interfaces import InterfaceClass, SetterProperty


class Interface(object, metaclass=InterfaceClass):
    """Interface class for testing interfaces."""

    def __init__(self):
        """Initialize members."""
        self.value = None

    @abc.abstractmethod
    def method(self):
        """Do nothing."""
        pass


class SecondInterface(object, metaclass=InterfaceClass):
    """Interface class for testing intefaces."""

    @abc.abstractmethod
    def method(self):
        pass


class Derived(Interface):
    """Derived class for testing interfaces."""

    def method(self):
        self.value = 0

    @SetterProperty
    def settable(self, value):
        """Modify the value member."""
        self.value = value


class SecondDerived(SecondInterface):
    """Derived class for testing interfaces."""

    def method(self):
        return True


def test_interface_class():
    """Test whether the InterfaceClass works for docstring inheritance."""
    test_object = Derived()
    assert test_object.value is None
    assert test_object.method.__doc__ == 'Do nothing.'
    test_object.method()
    assert test_object.value == 0

    test_object = SecondDerived()
    assert test_object.method()
    assert test_object.method.__doc__ == None


def test_setter_property():
    """Test whether the SetterProperty descriptor works for setting values."""
    test_object = Derived()
    assert test_object.value is None
    test_object.settable = 42
    assert test_object.value == 42



"""Test the pipelines module."""

# Builtins

# Packages

from phylline.links.clocked import DelayedEventLink
from phylline.links.events import EventLink
from phylline.links.links import ChunkedStreamLink
from phylline.links.loopback import TopLoopbackLink
from phylline.links.streams import StreamLink
from phylline.pipelines import AutomaticPipeline, ManualPipeline, PipelineBottomCoupler
from phylline.pipes import AutomaticPipe

import pytest

from tests.unit.links.links import HIGHER_CHUNKED_STREAM, LOWER_CHUNKED_STREAM
from tests.unit.pipes import (
    assert_bottom_events,
    write_bottom_chunked_buffers,
    write_top_events
)


def make_pipeline_short(pipeline_type):
    """Make a short pipeline."""
    pipeline = pipeline_type(
        ChunkedStreamLink(),
        EventLink(),
    )
    assert isinstance(pipeline.bottom, ChunkedStreamLink)
    assert isinstance(pipeline.top, EventLink)
    return pipeline


def make_pipeline_long(pipeline_type):
    """Make a long pipeline."""
    pipeline = pipeline_type(
        StreamLink(),
        StreamLink(),
        StreamLink(),
        ChunkedStreamLink(),
        EventLink(),
        EventLink(),
        EventLink()
    )
    assert isinstance(pipeline.bottom, StreamLink)
    assert isinstance(pipeline.top, EventLink)
    return pipeline


def make_pipeline_delayed(pipeline_type):
    """Make a short pipeline."""
    pipeline = pipeline_type(
        StreamLink(),
        ChunkedStreamLink(),
        DelayedEventLink(),
        EventLink(),
        EventLink()
    )
    assert isinstance(pipeline.bottom, StreamLink)
    assert isinstance(pipeline.top, EventLink)
    return pipeline


def make_pipeline_events(pipeline_type):
    """Make a events-only pipeline."""
    pipeline = pipeline_type(
        EventLink(),
        EventLink(),
        EventLink(),
        EventLink()
    )
    assert isinstance(pipeline.bottom, EventLink)
    assert isinstance(pipeline.top, EventLink)
    return pipeline


def make_pipeline_nested(pipeline_type):
    """Make a nested pipeline."""
    pipeline = pipeline_type(
        make_pipeline_long(pipeline_type),
        make_pipeline_events(pipeline_type),
        EventLink(),
        AutomaticPipe(EventLink(), EventLink()),
        make_pipeline_events(pipeline_type)
    )
    return pipeline


@pytest.mark.parametrize('pipeline_factory', [
    make_pipeline_short,
    make_pipeline_long,
    make_pipeline_nested
])
def test_manual_pipeline(pipeline_factory):
    """Exercise ManualPipeline's interface."""
    # print('Testing Manual Pipeline with factory {}:'.format(pipeline_factory.__name__))
    pipeline = pipeline_factory(ManualPipeline)
    # print(pipeline)

    # Read/write on links with directional sync
    write_bottom_chunked_buffers(pipeline.bottom)
    assert pipeline.sync_up() is None
    assert_bottom_events(pipeline.top)
    write_top_events(pipeline.top)
    assert pipeline.sync_down() is None
    result = pipeline.bottom.to_write()
    # print('Pipeline bottom wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM

    # Read/write on links with bidirectional sync
    write_bottom_chunked_buffers(pipeline.bottom)
    assert pipeline.sync() is None
    assert_bottom_events(pipeline.top)
    write_top_events(pipeline.top)
    assert pipeline.sync() is None
    result = pipeline.bottom.to_write()
    # print('Pipeline bottom wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM

    # Read/write on pipe with bidirectionl sync
    write_bottom_chunked_buffers(pipeline)
    assert pipeline.sync() is None
    assert_bottom_events(pipeline)
    write_top_events(pipeline)
    assert pipeline.sync() is None
    result = pipeline.to_write()
    # print('Pipeline bottom wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM


def test_manual_pipeline_clocked():
    """Exercise ManualPipeline's clock functionality."""
    # print('Testing Manual Pipeline with factory make_pipeline_delayed:')
    pipeline = make_pipeline_delayed(ManualPipeline)
    # print(pipeline)

    assert pipeline.update_clock(0) is None
    write_bottom_chunked_buffers(pipeline.bottom)
    assert pipeline.sync() == 1.0
    assert not pipeline.top.has_receive()
    assert pipeline.update_clock(0.5) == 1.0
    assert not pipeline.top.has_receive()
    assert pipeline.update_clock(0.99) == 1.0
    assert not pipeline.top.has_receive()
    assert pipeline.update_clock(1.0) is None
    assert_bottom_events(pipeline.top)

    # print('Resetting clock...')
    assert pipeline.update_clock(0) is None
    write_top_events(pipeline.top)
    assert pipeline.sync() == 1.0
    assert pipeline.update_clock(0.5) == 1.0
    assert not pipeline.bottom.to_write()
    assert pipeline.update_clock(0.75) == 1.0
    assert not pipeline.bottom.to_write()
    assert pipeline.update_clock(1.5) is None
    result = pipeline.bottom.to_write()
    assert result == HIGHER_CHUNKED_STREAM


@pytest.mark.parametrize('pipeline_factory', [
    make_pipeline_short,
    make_pipeline_long,
    make_pipeline_nested
])
def test_automatic_pipeline(pipeline_factory):
    """Exercise AutomaticPipeline's interface."""
    # print('Testing Automatic Pipeline with factory {}:'.format(pipeline_factory.__name__))
    automatic_pipeline = pipeline_factory(AutomaticPipeline)
    # print(automatic_pipeline)

    # Read/write on links
    write_bottom_chunked_buffers(automatic_pipeline.bottom)
    assert_bottom_events(automatic_pipeline.top)
    write_top_events(automatic_pipeline.top)
    result = automatic_pipeline.bottom.to_write()
    # print('Pipeline bottom wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM

    # Read/write on pipeline
    write_bottom_chunked_buffers(automatic_pipeline)
    assert_bottom_events(automatic_pipeline)
    write_top_events(automatic_pipeline)
    result = automatic_pipeline.to_write()
    # print('Pipeline bottom wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM


def test_automatic_pipeline_clocked():
    """Exercise AutomaticPipeline's clock functionality."""
    # print('Testing Automatic Pipeline with factory make_pipeline_delayed:')
    pipeline = make_pipeline_delayed(AutomaticPipeline)
    # print(pipeline)

    assert pipeline.update_clock(0) is None
    write_bottom_chunked_buffers(pipeline.bottom)
    assert pipeline.update_clock(0) == 1.0
    assert not pipeline.top.has_receive()
    assert pipeline.update_clock(0.5) == 1.0
    assert not pipeline.top.has_receive()
    assert pipeline.update_clock(0.99) == 1.0
    assert not pipeline.top.has_receive()
    assert pipeline.update_clock(1.0) is None
    assert_bottom_events(pipeline.top)

    # print('Resetting clock...')
    assert pipeline.update_clock(0) is None
    write_top_events(pipeline.top)
    assert pipeline.update_clock(0) == 1.0
    assert pipeline.update_clock(0.5) == 1.0
    assert not pipeline.bottom.to_write()
    assert pipeline.update_clock(0.75) == 1.0
    assert not pipeline.bottom.to_write()
    assert pipeline.update_clock(1.5) is None
    result = pipeline.bottom.to_write()
    assert result == HIGHER_CHUNKED_STREAM


def make_pipeline_loopback(pipeline_factory):
    """Make a long pipeline with a loopback at the top."""
    manual_pipeline = pipeline_factory(
        StreamLink(),
        StreamLink(),
        StreamLink(),
        ChunkedStreamLink(),
        EventLink(),
        EventLink(),
        EventLink(),
        TopLoopbackLink()
    )
    assert isinstance(manual_pipeline.bottom, StreamLink)
    assert isinstance(manual_pipeline.top, TopLoopbackLink)
    return manual_pipeline


def test_manual_pipeline_loopback():
    """Exercise ManualPipeline's interface."""
    # print('Testing Manual Pipeline with factory make_pipeline_loopback:')
    manual_pipeline = make_pipeline_loopback(ManualPipeline)
    # print(manual_pipeline)

    write_bottom_chunked_buffers(manual_pipeline.bottom)
    assert manual_pipeline.sync() is None
    result = manual_pipeline.bottom.to_write()
    # print('Pipeline bottom wrote to stream: {}'.format(result))
    assert result == LOWER_CHUNKED_STREAM


def test_automatic_pipeline_loopback():
    """Exercise ManualPipeline's interface."""
    # print('Testing Automatic Pipeline with factory make_pipeline_loopback:')
    automatic_pipeline = make_pipeline_loopback(AutomaticPipeline)
    # print(automatic_pipeline)

    write_bottom_chunked_buffers(automatic_pipeline.bottom)
    result = automatic_pipeline.bottom.to_write()
    # print('Pipeline bottom wrote to stream: {}'.format(result))
    assert result == LOWER_CHUNKED_STREAM


def assert_loopback_below(stack, payload):
    """Asset that the stack has correct below-loopback behavior."""
    stack.send(payload)
    assert stack.has_receive()
    result = stack.receive()
    # print('Loopback received: {}'.format(result))
    assert result.data == payload


def test_loopback_pipeline_bottom_coupler():
    """Test for pipeline bottom coupling."""
    # print('Testing byte buffer loopback with PipelineBottomCoupler...')
    pipeline_one = make_pipeline_nested(AutomaticPipeline)
    pipeline_two = make_pipeline_loopback(AutomaticPipeline)
    coupler = PipelineBottomCoupler(pipeline_one, pipeline_two)
    # print(coupler)
    payload = b'\1\2\3\4'
    assert_loopback_below(pipeline_one.top, payload)
    assert_loopback_below(pipeline_one, payload)
    assert_loopback_below(coupler.pipeline_one, payload)



"""Test the links.events module."""

# Builtins

# Packages

from phylline.links.links import ChunkedStreamLink

from tests.unit.links.streams import HIGHER_BUFFERS, LOWER_BUFFERS


LOWER_CHUNKED_BUFFERS = [b'\0' + event + b'\0' for event in LOWER_BUFFERS]
LOWER_CHUNKED_STREAM = ''.join(
    buffer.decode('utf-8') for buffer in LOWER_CHUNKED_BUFFERS
).encode('utf-8')
HIGHER_CHUNKED_BUFFERS = [b'\0' + event + b'\0' for event in HIGHER_BUFFERS]
HIGHER_CHUNKED_BUFFERS_MINIMAL = [event + b'\0' for event in HIGHER_BUFFERS]
HIGHER_CHUNKED_STREAM = ''.join(
    buffer.decode('utf-8') for buffer in HIGHER_CHUNKED_BUFFERS
).encode('utf-8')
HIGHER_CHUNKED_STREAM_MINIMAL = ''.join(
    buffer.decode('utf-8') for buffer in HIGHER_CHUNKED_BUFFERS_MINIMAL
).encode('utf-8')


def test_chunked_stream_link():
    """Exercise ChunkedStreamLink's interface."""
    # print('Testing Chunked Stream Link:')
    chunked_stream_link = ChunkedStreamLink()
    chunked_stream_link.to_read(LOWER_CHUNKED_BUFFERS[0])
    chunked_stream_link.to_read(LOWER_CHUNKED_BUFFERS[1] + LOWER_CHUNKED_BUFFERS[2])
    assert chunked_stream_link.has_receive()
    for (i, event) in enumerate(chunked_stream_link.receive_all()):
        # print('Chunked Stream Link received from stream: {}'.format(event))
        assert event.data == LOWER_BUFFERS[i]
    for event in HIGHER_BUFFERS:
        chunked_stream_link.send(event)
    result = chunked_stream_link.to_write()
    # print('Chunked Stream Link wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM

    # print('Testing Chunked Stream Link without chunk starting delimiters:')
    chunked_stream_link = ChunkedStreamLink(begin_chunk_separator=False)
    chunked_stream_link.to_read(LOWER_CHUNKED_BUFFERS[0])
    chunked_stream_link.to_read(LOWER_CHUNKED_BUFFERS[1] + LOWER_CHUNKED_BUFFERS[2])
    assert chunked_stream_link.has_receive()
    for (i, event) in enumerate(chunked_stream_link.receive_all()):
        # print('Chunked Stream Link received from stream: {}'.format(event))
        assert event.data == LOWER_BUFFERS[i]
    for event in HIGHER_BUFFERS:
        chunked_stream_link.send(event)
    result = chunked_stream_link.to_write()
    # print('Chunked Stream Link wrote to stream: {}'.format(result))
    assert result == HIGHER_CHUNKED_STREAM_MINIMAL



"""setup.py for phyllo package."""
import subprocess

import setuptools
from setuptools.command.test import test as testcommand


PACKAGE_NAME = 'phylline'


# CUSTOM TEST COMMANDS


class TestQuickCommand(testcommand):
    """setup.py command to run a quick test prioritizing unstaged changes."""

    description = 'Quick test for unstaged changes and previous failures.'

    def run_tests(self):
        """Run quick unstaged change test."""
        subprocess.check_call([
            'pytest',
            '-n', 'auto',
            '--picked=first',
            '--failed-first'
        ])


class TestQuickBranchCommand(testcommand):
    """setup.py command to run a quick test prioritizing branch changes."""

    description = 'Quick test for branch changes and previous failures.'

    def run_tests(self):
        """Run quick branch change test."""
        subprocess.check_call([
            'pytest',
            '-n', 'auto',
            '--picked=first',
            '--mode=branch',
            '--failed-first'
        ])


class TestDebugCommand(testcommand):
    """setup.py command to run a test and switch to pdb on the first failure."""

    description = 'Quick test for pdb on the first failure.'

    def run_tests(self):
        """Run debugging test."""
        subprocess.check_call([
            'pytest',
            '-n', 'auto',
            '--picked=first',
            '--failed-first',
            '--pdb', '-x'
        ])


class TestPerformanceCommand(testcommand):
    """setup.py command to run a performance profiling test."""

    description = 'Test with performance profiling.'

    def run_tests(self):
        """Run performance profiling test."""
        subprocess.check_call([
            'pytest',
            '--profile', '--profile-svg'
        ])


class TestCoverageCommand(testcommand):
    """setup.py command to run a coverage reporting test."""

    description = 'Test with coverage reporting.'

    def run_tests(self):
        """Run coverage reporting test."""
        subprocess.check_call([
            'pytest',
            '--cov={}'.format(PACKAGE_NAME),
            '--cov-branch',
            '--cov-report', 'term-missing:skip-covered',
            '--cov-report', 'html'
        ])


class TestCompleteCommand(testcommand):
    """setup.py command to run a complete reporting test."""

    description = 'Test with complete reporting.'

    def run_tests(self):
        """Run coverage reporting test."""
        subprocess.check_call([
            'pytest',
            '--durations=0',
            '--cov={}'.format(PACKAGE_NAME),
            '--cov-branch',
            '--cov-report', 'term-missing',
            '--cov-report', 'html',
            '--hypothesis-show-statistics',
            '--verbose'
        ])


# MAIN SETUP

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name=PACKAGE_NAME,
    version='0.1.0',
    description=(
        'Sans-I/O bidirectional data pipeline framework for making layered '
        'communication protocols.',
    ),
    url='https://github.com/ethanjli/phylline',
    author='Ethan Li',
    author_email='lietk12@gmail.com',
    packages=setuptools.find_packages(),
    install_requires=[
        'ohneio'
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'hypothesis'],
    cmdclass={
        'test_quick': TestQuickCommand,
        'test_quick_branch': TestQuickBranchCommand,
        'test_debug': TestDebugCommand,
        'test_perf': TestPerformanceCommand,
        'test_coverage': TestCoverageCommand,
        'test_complete': TestCompleteCommand
    }
)



"""Utilities for working with iterables."""

# Builtins

# Packages


def make_collection(iterable_or_singleton, type=list):
    """Make a collection from either an iterable of elements or a singleton element."""
    try:
        iterator = iter(iterable_or_singleton)
        return type(iterator)
    except TypeError:
        return type((iterable_or_singleton,))


def remove_none(iterable):
    """Return all Nones from an iterable."""
    return (item for item in iterable if item is not None)
